/* Keen Dreams Source Code
 * Copyright (C) 2014 Javier M. Chavez
 * Copyright (C) 1993-2014 Flat Rock Software
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

// KD_MISC.C

#include "KD_DEF.H"
#pragma	hdrstop

/*
=============================================================================

LOCAL CONSTANTS

=============================================================================
*/


/*
=============================================================================

GLOBAL VARIABLES

=============================================================================
*/


/*
=============================================================================

LOCAL VARIABLES

=============================================================================
*/


//===========================================================================
//
// LARVAC (Slug replacement)
// 
//===========================================================================

CK_action AR_ACT_LarvacWalk0;
CK_action AR_ACT_LarvacWalk1;
CK_action AR_ACT_LarvacSliming;
CK_action AR_ACT_LarvacStunned;
CK_action AR_ACT_LarvacGoop1;
CK_action AR_ACT_LarvacGoop2;

void AR_SpawnLarvac(int tileX, int tileY);
void AR_LarvacWalk(CK_object *obj);
void AR_LarvacGoop(CK_object *obj);
void AR_LarvacCollide(CK_object *larvac, CK_object *obj2);
void AR_LarvacGoopReact(CK_object *obj);


CK_action AR_ACT_LarvacWalk0 = {SPR_LARVACWALK1, SPR_LARVACWALK2, AT_UnscaledOnce, 0, 1, 8, 64, 0, AR_LarvacWalk, AR_LarvacCollide, CK_BasicDrawFunc2, &AR_ACT_LarvacWalk1};
CK_action AR_ACT_LarvacWalk1 = {SPR_LARVACWALK2, SPR_LARVACWALK1, AT_UnscaledOnce, 0, 1, 8, 64, 0, AR_LarvacWalk, AR_LarvacCollide, CK_BasicDrawFunc2, &AR_ACT_LarvacWalk0};

CK_action AR_ACT_LarvacSliming = {SPR_LARVACEXCR, SPR_LARVACEXCR, AT_UnscaledOnce, 0, 1, 70, 0, 0, AR_LarvacGoop, AR_LarvacCollide, CK_BasicDrawFunc2, &AR_ACT_LarvacWalk0};

CK_action AR_ACT_LarvacStunned = {SPR_LARVACDEAD, SPR_LARVACDEAD, AT_Frame, 0, 0, 0, 0, 0, CK_Fall, NULL, CK_BasicDrawFunc4, NULL};

CK_action AR_ACT_LarvacGoop1 = {SPR_LARSLIME1, SPR_LARSLIME1, AT_ScaledOnce, 0, 0, 8, 0, 10, NULL, CK_DeadlyCol2, AR_LarvacGoopReact, &AR_ACT_LarvacGoop2};
CK_action AR_ACT_LarvacGoop2 = {SPR_LARSLIME2, SPR_LARSLIME2, AT_ScaledOnce, 0, 0, 8, 0, 10, NULL, CK_DeadlyCol2, AR_LarvacGoopReact, &AR_ACT_LarvacGoop1};


// Note: Spawn actor slightly above ground so that
// clipping function pushes it on to the ground, as stick to ground
// option is set in action
void AR_SpawnLarvac(int tileX, int tileY)
{
	CK_object *new_object = CK_GetNewObj(false);
	new_object->type = CT_Larvac;
	new_object->active = OBJ_ACTIVE;
	new_object->zLayer = 0;
	new_object->posX = tileX << 8;
	new_object->posY = (tileY << 8) - 0x80; 
	new_object->xDirection = US_RndT() < 0x80 ? motion_Left : motion_Right;
	new_object->yDirection = motion_None;
	new_object->clipped = CLIP_normal;
	CK_SetAction(new_object, &AR_ACT_LarvacWalk0);
}

void AR_LarvacWalk(CK_object *obj)
{
	if (US_RndT() < 0x10)
	{
		CK_SetAction2(obj, &AR_ACT_LarvacSliming);
		return;
	}

}

void AR_LarvacGoop(CK_object *obj)
{
	// Spawn a larvac slime
	CK_object *new_object = CK_GetNewObj(false);
	new_object->type = CT_LarvacGoop;
	new_object->active = OBJ_EXISTS_ONLY_ONSCREEN;
	new_object->zLayer = 0;
	new_object->posX = obj->posX;
	new_object->posY = obj->posY;
	new_object->xDirection = motion_None;
	new_object->yDirection = motion_Up;
	new_object->clipped = CLIP_normal;
	CK_SetAction(new_object, &AR_ACT_LarvacGoop1);
	SD_PlaySound(SLUGPOOSND);
}
	
void AR_LarvacCollide(CK_object *obj1, CK_object *obj2)
{

	if (obj2->type == CT_Player)
	{
		CK_KillKeen();
	}
	else if (obj2->type == CT_Stunner)
	{
		CK_StunCreature(obj1, obj2, &AR_ACT_LarvacStunned);
	}
}


// The slime dissipates when it touches a surface
void AR_LarvacGoopReact(CK_object *obj)
{
	if (obj->topTI || obj->bottomTI || obj->leftTI || obj->rightTI)
	{
		CK_RemoveObj(obj);
		return;
	}

	RF_PlaceSprite(&obj->sde, obj->posX, obj->posY, obj->gfxChunk, false, obj->zLayer);

}


//===========================================================================
//
// BEETLEBONE (Mad Mushroom Replacement)
// 
//===========================================================================

// Deadly on contact; starts bouncing up and down when keen within 3 tiles

CK_action AR_ACT_BeetleWait;
CK_action AR_ACT_BeetleIdle1;
CK_action AR_ACT_BeetleIdle2;
CK_action AR_ACT_BeetleJump1;
CK_action AR_ACT_BeetleJump2;

void AR_SpawnBeetle(int tileX, int tileY);
void AR_BeetleIdle (CK_object *obj);
void AR_BeetleFallReact (CK_object *obj);

CK_action AR_ACT_BeetleWait = {SPR_BEETLEIDLE1, SPR_BEETLEIDLE1, AT_UnscaledOnce, 0, 0, 35, 0, 0, NULL, CK_DeadlyCol, CK_BasicDrawFunc1, &AR_ACT_BeetleIdle1};

CK_action AR_ACT_BeetleIdle1 = {SPR_BEETLEIDLE1, SPR_BEETLEIDLE1, AT_UnscaledFrame, 0, 0, 8, 0, 0, AR_BeetleIdle, CK_DeadlyCol, CK_BasicDrawFunc1, &AR_ACT_BeetleIdle2};
CK_action AR_ACT_BeetleIdle2 = {SPR_BEETLEIDLE2, SPR_BEETLEIDLE2, AT_UnscaledFrame, 0, 0, 8, 0, 0, AR_BeetleIdle, CK_DeadlyCol, CK_BasicDrawFunc1, &AR_ACT_BeetleIdle1};


CK_action AR_ACT_BeetleJump1 = {SPR_BEETLEJUMP1, SPR_BEETLEJUMP1, AT_UnscaledFrame, 0, 0, 12, 0, 0, CK_Fall, CK_DeadlyCol, AR_BeetleFallReact, &AR_ACT_BeetleJump2};
CK_action AR_ACT_BeetleJump2 = {SPR_BEETLEJUMP2, SPR_BEETLEJUMP2, AT_UnscaledFrame, 0, 0, 12, 0, 0, CK_Fall, CK_DeadlyCol, AR_BeetleFallReact, &AR_ACT_BeetleJump1};


void AR_SpawnBeetle(int tileX, int tileY)
{
	CK_object *new_object = CK_GetNewObj(false);
	new_object->type = CT_Beetle;
	new_object->active = OBJ_ACTIVE;
	new_object->zLayer = 0;
	new_object->posX = (tileX << 8) - 0;
	new_object->posY = (tileY << 8) - 0x100; 
	new_object->xDirection = motion_None;
	new_object->yDirection = motion_None;
	new_object->clipped = CLIP_normal;
	CK_SetAction(new_object, &AR_ACT_BeetleIdle1);
}

void AR_BeetleIdle (CK_object *obj)
{
	int deltaX = ck_keenObj->posX - obj->posX;
	int deltaY = ck_keenObj->posY - obj->posY;

	// Wait for player
	if (deltaX < 0x300 && deltaX > -0x300 && deltaY < 0)
	{
		obj->velY = -80;
		obj->currentAction = &AR_ACT_BeetleJump1;
		SD_PlaySound(BEETLEJUMPSND);
	}	
}


void AR_BeetleFallReact (CK_object *obj)
{
	if (obj->topTI)
		CK_SetAction2(obj, &AR_ACT_BeetleWait);

	RF_PlaceSprite(&obj->sde, obj->posX, obj->posY, obj->gfxChunk, false, obj->zLayer);
}



//===========================================================================
//
// MAGMINE (Inchworm replacement)
// 
//===========================================================================

// Description: Harmless, follow keen.  When two collide, spawn deadly explosion actor of transient existence.

CK_action AR_ACT_MagmineWalking1;
CK_action AR_ACT_MagmineWalking2;
CK_action AR_ACT_MagmineWalking3;
CK_action AR_ACT_MagmineWalking4;

CK_action AR_ACT_MagmineExplosion1;
CK_action AR_ACT_MagmineExplosion2;
CK_action AR_ACT_MagmineExplosion3;
CK_action AR_ACT_MagmineExplosion4;
CK_action AR_ACT_MagmineExplosion5;

void AR_SpawnMagmine(int tileX, int tileY);
void AR_MagmineWalk(CK_object *obj);
void AR_MagmineCol(CK_object *obj1, CK_object *obj2);
void AR_MagmineReact(CK_object *obj);

CK_action AR_ACT_MagmineWalking1 = {SPR_MAGMINE1, SPR_MAGMINE1, AT_ScaledFrame, 1, 1, 8, 10, 0, AR_MagmineWalk, AR_MagmineCol, AR_MagmineReact, &AR_ACT_MagmineWalking2};
CK_action AR_ACT_MagmineWalking2 = {SPR_MAGMINE2, SPR_MAGMINE2, AT_ScaledFrame, 1, 1, 8, 10, 0, AR_MagmineWalk, AR_MagmineCol, AR_MagmineReact, &AR_ACT_MagmineWalking3};
CK_action AR_ACT_MagmineWalking3 = {SPR_MAGMINE3, SPR_MAGMINE3, AT_ScaledFrame, 1, 1, 8, 10, 0, AR_MagmineWalk, AR_MagmineCol, AR_MagmineReact, &AR_ACT_MagmineWalking4};
CK_action AR_ACT_MagmineWalking4 = {SPR_MAGMINE4, SPR_MAGMINE4, AT_ScaledFrame, 1, 1, 8, 10, 0, AR_MagmineWalk, AR_MagmineCol, AR_MagmineReact, &AR_ACT_MagmineWalking1};

CK_action AR_ACT_MagmineExplosion1 = {SPR_MAGMINEEXPL1, SPR_MAGMINEEXPL1, AT_UnscaledFrame, 0, 0, 8, 0, 0, NULL, CK_DeadlyCol2, CK_BasicDrawFunc1, &AR_ACT_MagmineExplosion2};
CK_action AR_ACT_MagmineExplosion2 = {SPR_MAGMINEEXPL2, SPR_MAGMINEEXPL2, AT_UnscaledFrame, 0, 0, 8, 0, 0, NULL, CK_DeadlyCol2, CK_BasicDrawFunc1, &AR_ACT_MagmineExplosion3};
CK_action AR_ACT_MagmineExplosion3 = {SPR_MAGMINEEXPL3, SPR_MAGMINEEXPL3, AT_UnscaledFrame, 0, 0, 8, 0, 0, NULL, CK_DeadlyCol2, CK_BasicDrawFunc1, &AR_ACT_MagmineExplosion4};
CK_action AR_ACT_MagmineExplosion4 = {SPR_MAGMINEEXPL4, SPR_MAGMINEEXPL4, AT_UnscaledFrame, 0, 0, 8, 0, 0, NULL, CK_DeadlyCol2, CK_BasicDrawFunc1, &AR_ACT_MagmineExplosion5};
CK_action AR_ACT_MagmineExplosion5 = {SPR_MAGMINEEXPL5, SPR_MAGMINEEXPL5, AT_UnscaledFrame, 0, 0, 8, 0, 0, NULL, CK_DeadlyCol2, CK_BasicDrawFunc1, NULL};



void AR_SpawnMagmine(int tileX, int tileY)
{
	CK_object *new_object = CK_GetNewObj(false);
	new_object->type = CT_Magmine;
	new_object->active = OBJ_ACTIVE;
	new_object->zLayer = 0;
	new_object->posX = tileX << 8;
	new_object->posY = (tileY << 8); 
	new_object->xDirection = US_RndT() < 0x80 ? motion_Left : motion_Right;
	new_object->yDirection = motion_None;
	new_object->clipped = CLIP_normal;
	CK_SetAction(new_object, &AR_ACT_MagmineWalking1);
}

void AR_MagmineWalk(CK_object *obj)
{
	// Follow Keen 
	if (!obj->xDirection)
		obj->xDirection = obj->clipRects.unitXmid>ck_keenObj->clipRects.unitXmid?motion_Left:motion_Right;


	if (obj->posX < ck_keenObj->posX - 0x80)
		obj->xDirection = motion_Right;

	if (obj->posX > ck_keenObj->posX + 0x80)
		obj->xDirection = motion_Left;

}

void AR_MagmineCol(CK_object *obj1, CK_object *obj2)
{
	if (obj2->type == CT_Magmine)
	{
		CK_object *leftmost = obj1->posX < obj2->posX? obj1 : obj2;

		// Spawn an explostion sprite
		CK_object *new_object = CK_GetNewObj(false);
		new_object->type = CT_MagmineExplosion;
		new_object->active = OBJ_EXISTS_ONLY_ONSCREEN;
		new_object->zLayer = 2;
		new_object->posX = leftmost->posX;
		new_object->posY = leftmost->posY-0x100;
		new_object->xDirection = motion_None;
		new_object->yDirection = motion_None;
		new_object->clipped = CLIP_not;
		CK_SetAction(new_object, &AR_ACT_MagmineExplosion1);
		SD_PlaySound(MAGMINEEXPLODESND);

		// Remove both objects
		obj1->type = obj2->type = CT_nothing;
		CK_RemoveObj(obj1);
		CK_RemoveObj(obj2);
		return;
	}
}

void AR_MagmineReact (CK_object *obj)
{


	// IF a bridge was removed from underneath, then just hover
	if (!obj->topTI)
	{
		int x,y;
		for (y = obj->clipRects.tileY1; y <= obj->clipRects.tileY2 + 1; y++)
		{
			x = obj->clipRects.tileXmid;
			//for (x = obj->clipRects.tileX1; x <= obj->clipRects.tileX2; x++)
			{
				unsigned tile = *((unsigned _seg *)mapsegs[1] + mapbwidthtable[y]/2 + x);
				if (tinf[tile + NORTHWALL])
				{
					goto normalDraw;
				}
			}
		}

		obj->posX -= obj->deltaPosX;
		obj->posY -= obj->deltaPosY;
		CK_BasicDrawFunc1(obj);
		return;
	}

normalDraw:
	CK_BasicDrawFunc2(obj);


}


//===========================================================================
//
// FIREBALL (Mine replacement)
// 
//===========================================================================

// Description: Fireball moves back and forth, deadly to touch.

CK_action AR_ACT_Fireball1;
CK_action AR_ACT_Fireball2;
CK_action AR_ACT_Fireball3;

CK_action AR_ACT_Fireball1 = {SPR_FIREBALL1, SPR_FIREBALL1, AT_ScaledFrame, 0, 0, 16, 0, 0, CK5_RedAxisPlatform, CK_DeadlyCol, CK_BasicDrawFunc1, &AR_ACT_Fireball2 };
CK_action AR_ACT_Fireball2 = {SPR_FIREBALL2, SPR_FIREBALL2, AT_ScaledFrame, 0, 0, 16, 0, 0, CK5_RedAxisPlatform, CK_DeadlyCol, CK_BasicDrawFunc1, &AR_ACT_Fireball3 };
CK_action AR_ACT_Fireball3 = {SPR_FIREBALL3, SPR_FIREBALL3, AT_ScaledFrame, 0, 0, 16, 0, 0, CK5_RedAxisPlatform, CK_DeadlyCol, CK_BasicDrawFunc1, &AR_ACT_Fireball1 };

void AR_SpawnFireball(int tileX, int tileY, int direction)
{
	CK_object *obj = CK_GetNewObj(false);

	obj->type = CT_Fireball;
	obj->active = OBJ_ALWAYS_ACTIVE;
	obj->zLayer = 0;
	obj->posX = tileX << 8;
	obj->posY = tileY << 8;
	obj->clipped = CLIP_not;

	switch (direction)
	{
	case 0:
		obj->xDirection = 1;
		obj->yDirection = 0;
		break;
	case 1:
		obj->xDirection = 0;
		obj->yDirection = 1;
		break;
	case 2:
		obj->xDirection = -1;
		obj->yDirection = 0;
		break;
	}

	CK_SetAction(obj, &AR_ACT_Fireball1);
	//obj->gfxChunk = obj->currentAction->chunkLeft;

	CK_ResetClipRects(obj);
}


//===========================================================================
//
// DARTFROG (Bounder replacement)
// 
//===========================================================================

CK_action AR_ACT_DartFrogIdle1;
CK_action AR_ACT_DartFrogJump1;
CK_action AR_ACT_DartFrogStunned;
CK_action AR_ACT_DartFrogFunny1;
CK_action AR_ACT_DartFrogFunny2;
CK_action AR_ACT_DartFrogFunny3;

void AR_SpawnDartfrog(int tileX, int tileY);
void AR_DartFrogIdle(CK_object *obj);
void AR_DartFrogCol(CK_object *obj1, CK_object *obj2);
void AR_DartFrogJumpReact(CK_object *obj);

// Description: Hops along ground toward keen, pausing in between hops.  Deadly to touch, takes 1 shot to stun.
CK_action AR_ACT_DartFrogIdle1 = {SPR_DARTFROGSITL, SPR_DARTFROGSITR, AT_UnscaledOnce, 0, 0, 60, 0, 0, AR_DartFrogIdle, AR_DartFrogCol, CK_BasicDrawFunc1, &AR_ACT_DartFrogJump1};
CK_action AR_ACT_DartFrogJump1 = {SPR_DARTFROGJUMPL, SPR_DARTFROGJUMPR, AT_Frame, 0, 0, 0, 0, 0, CK_Fall, AR_DartFrogCol, AR_DartFrogJumpReact, &AR_ACT_DartFrogJump1};

CK_action AR_ACT_DartFrogStunned = {SPR_DARTFROGDEAD, SPR_DARTFROGDEAD, AT_Frame, 0, 0, 0, 0, 0, CK_Fall, NULL, CK_BasicDrawFunc4, NULL};

CK_action AR_ACT_DartFrogFunny1 = {SPR_DARTFROGIDLE, SPR_DARTFROGIDLE, AT_UnscaledOnce, 0, 0, 70, 0, 0, NULL, AR_DartFrogCol, CK_BasicDrawFunc1, &AR_ACT_DartFrogFunny2};
CK_action AR_ACT_DartFrogFunny2 = {SPR_DARTFROGIDLE2, SPR_DARTFROGIDLE2, AT_UnscaledOnce, 0, 0, 25, 0, 0, NULL, AR_DartFrogCol, CK_BasicDrawFunc1, &AR_ACT_DartFrogFunny3};
CK_action AR_ACT_DartFrogFunny3 = {SPR_DARTFROGIDLE, SPR_DARTFROGIDLE, AT_UnscaledOnce, 0, 0, 35, 0, 0, NULL, AR_DartFrogCol, CK_BasicDrawFunc1, &AR_ACT_DartFrogIdle1};

void AR_SpawnDartfrog(int tileX, int tileY)
{
	CK_object *new_object = CK_GetNewObj(false);
	new_object->type = CT_DartFrog;
	new_object->active = OBJ_ACTIVE;
	new_object->zLayer = 0;
	new_object->posX = (tileX << G_T_SHIFT);
	new_object->posY = (tileY << G_T_SHIFT) - 0x80; 
	new_object->xDirection = new_object->posX > ck_keenObj->posX? motion_Left : motion_Right;
	new_object->yDirection = motion_None;
	new_object->clipped = CLIP_normal;
	CK_SetAction(new_object, &AR_ACT_DartFrogIdle1);

	if (!new_object->topTI)
		CK_SetAction2(new_object, &AR_ACT_DartFrogJump1);
}

void AR_DartFrogIdle(CK_object *obj)
{
	
	if (obj->user1++ >= 3 && US_RndT() < 0x80)
	{
		obj->user1 = 0;
		CK_SetAction2(obj, &AR_ACT_DartFrogFunny1);
		return;
	}
		
	// Face Keen
	if (ck_keenObj->posX < obj->posX)
		obj->xDirection = motion_Left;
	else
		obj->xDirection = motion_Right;

	// Jump
	obj->velY = -50;
	obj->velX = obj->xDirection * 16;
	SD_PlaySound(DARTFROGJUMPSND);
	CK_SetAction2(obj, &AR_ACT_DartFrogJump1);

}

void AR_DartFrogCol(CK_object *obj1, CK_object *obj2)
{
	if (obj2->type == CT_Player)
	{
		CK_KillKeen();
	}
	else if (obj2->type == CT_Stunner)
	{
		CK_StunCreature(obj1, obj2, &AR_ACT_DartFrogStunned);
	}
}

void AR_DartFrogJumpReact(CK_object *obj)
{
	if (obj->topTI)
	{
		CK_SetAction2(obj, &AR_ACT_DartFrogIdle1);
		obj->velX = obj->velY = 0;
	}

	if (obj->leftTI || obj->rightTI)
	{
		obj->xDirection = -obj->xDirection;
		obj->velX = -obj->velX;
	}
	


	RF_PlaceSprite(&obj->sde, obj->posX, obj->posY, obj->gfxChunk, false, obj->zLayer);
	
}

//===========================================================================
//
// Splider (Mimrock replacement)
// 
//===========================================================================

// Description: Walks around, invulnerable.  Slides at keen when close, exposing it's beak
// to stunner susceptibility.

CK_action AR_ACT_SpliderIdle;

CK_action AR_ACT_SpliderWalk1;
CK_action AR_ACT_SpliderWalk2;
CK_action AR_ACT_SpliderWalk3;
CK_action AR_ACT_SpliderWalk4;

CK_action AR_ACT_SpliderSlide1;
CK_action AR_ACT_SpliderSlide2;

CK_action AR_ACT_SpliderAbortSlide;
CK_action AR_ACT_SpliderStunned;

void AR_SpawnSplider(int tileX, int tileY);
void AR_SpliderWalk(CK_object *obj);
void AR_SpliderSearch(CK_object *obj);
void AR_SpliderSlideEnd(CK_object *obj);
void AR_SpliderSlideReact(CK_object *obj);
void AR_SpliderFallingSlideReact(CK_object *obj);
void AR_SpliderAbortSlideReact(CK_object *obj);
void AR_SpliderSlideCol(CK_object *obj1, CK_object *obj2);


CK_action AR_ACT_SpliderIdle = {SPR_SPLIDERIDLE, SPR_SPLIDERIDLE, AT_UnscaledOnce, 0, 1, 35, 0, 0, AR_SpliderSearch, CK_DeadlyCol, CK_BasicDrawFunc1, &AR_ACT_SpliderWalk1};

CK_action AR_ACT_SpliderWalk1 = {SPR_SPLIDERWALKL1, SPR_SPLIDERWALKR1, AT_UnscaledOnce, 0, 1, 6, 64, 0, AR_SpliderWalk, CK_DeadlyCol, CK_BasicDrawFunc2, &AR_ACT_SpliderWalk2};
CK_action AR_ACT_SpliderWalk2 = {SPR_SPLIDERWALKL2, SPR_SPLIDERWALKR2, AT_UnscaledOnce, 0, 1, 6, 64, 0, AR_SpliderWalk, CK_DeadlyCol, CK_BasicDrawFunc2, &AR_ACT_SpliderWalk3};
CK_action AR_ACT_SpliderWalk3 = {SPR_SPLIDERWALKL3, SPR_SPLIDERWALKR3, AT_UnscaledOnce, 0, 1, 6, 64, 0, AR_SpliderWalk, CK_DeadlyCol, CK_BasicDrawFunc2, &AR_ACT_SpliderWalk4};
CK_action AR_ACT_SpliderWalk4 = {SPR_SPLIDERWALKL4, SPR_SPLIDERWALKR4, AT_UnscaledOnce, 0, 1, 6, 64, 0, AR_SpliderWalk, CK_DeadlyCol, CK_BasicDrawFunc2, &AR_ACT_SpliderWalk1};

CK_action AR_ACT_SpliderSlide1 = {SPR_SPLIDERSLIDEL1, SPR_SPLIDERSLIDER1, AT_ScaledFrame, 0, 1, 48, 0, 0, CK_Fall2, AR_SpliderSlideCol, AR_SpliderSlideReact, &AR_ACT_SpliderSlide2};
CK_action AR_ACT_SpliderSlide2 = {SPR_SPLIDERSLIDEL1, SPR_SPLIDERSLIDER1, AT_Frame, 0, 1, 6, 0, 0, AR_SpliderSlideEnd, AR_SpliderSlideCol, AR_SpliderSlideReact, &AR_ACT_SpliderSlide2};

CK_action AR_ACT_SpliderFallingSlide = {SPR_SPLIDERSLIDEL1, SPR_SPLIDERSLIDER1, AT_Frame, 0, 0, 6, 0, 0, AR_SpliderSlideEnd, AR_SpliderSlideCol, AR_SpliderFallingSlideReact, &AR_ACT_SpliderFallingSlide};


CK_action AR_ACT_SpliderAbortSlide = {SPR_SPLIDERSLIDEL1, SPR_SPLIDERSLIDER1, AT_Frame, 0, 0, 48, 0, 0, CK_Fall2, AR_SpliderSlideCol, AR_SpliderAbortSlideReact, &AR_ACT_SpliderAbortSlide};

CK_action AR_ACT_SpliderStunned = {SPR_SPLIDERDEAD, SPR_SPLIDERDEAD, AT_Frame, 0, 0, 0, 0, 0, CK_Fall, NULL, CK_BasicDrawFunc4, NULL};

void AR_SpawnSplider(int tileX, int tileY)
{
	CK_object *new_object = CK_GetNewObj(false);
	new_object->type = CT_Splider;
	new_object->active = OBJ_ACTIVE;
	new_object->zLayer = PRIORITIES - 4;
	new_object->posX = tileX << 8;
	new_object->posY = (tileY << 8) - 0x100; 
	new_object->xDirection = US_RndT() < 0x80 ? motion_Left : motion_Right;
	new_object->yDirection = motion_None;
	new_object->clipped = CLIP_normal;
	CK_SetAction(new_object, &AR_ACT_SpliderWalk1);
}


void AR_SpliderWalk(CK_object *obj)
{
	unsigned distanceToKeenX;
	int distanceToKeenY;

	// Turn around randomly 
	if (US_RndT() < 0x4)
	{
		CK_SetAction2(obj, &AR_ACT_SpliderIdle);
		return;
	}


	// Slide attack if Keen is within range
	distanceToKeenX = ck_keenObj->clipRects.unitXmid - obj->clipRects.unitXmid;
	distanceToKeenY = ck_keenObj->posY - obj->posY;

	if ((ABS(distanceToKeenY) < 0x300) &&
	 ((obj->xDirection == motion_Right && distanceToKeenX < 0x500) ||
	 obj->xDirection == motion_Left && distanceToKeenX > (unsigned)-0x500))
	{
		SD_PlaySound(SPLIDERSND);
		CK_SetAction2(obj, &AR_ACT_SpliderSlide1);
		obj->velX = obj->xDirection * 30;
		return;
	}
	
}

void AR_SpliderSearch(CK_object *obj)
{

	if (obj->posX < ck_keenObj->posX)
	{
		obj->xDirection = motion_Right;
	}
	else
	{
		obj->xDirection = motion_Left;
	}

}

void AR_SpliderSlideEnd(CK_object *obj)
{
	// TODO: Play a sound when the decelleration starts
	
	// Glide to a stop
	CK_PhysDampHorz(obj);
	CK_PhysGravityMid(obj);



}

void AR_SpliderSlideReact(CK_object *obj)
{
	if (obj->xDirection == motion_Right && obj->leftTI || obj->xDirection == motion_Left && obj->rightTI)
	{
		obj->velX = 0;
		CK_SetAction2(obj, &AR_ACT_SpliderAbortSlide);
	}

	if (obj->topTI)
	{
		obj->velY = 0;
		if (obj->velX == 0)
			CK_SetAction2(obj, &AR_ACT_SpliderIdle);
	}
	else
	{
		// Splider's fallen off a ledge; switch to slide-while-falling
		obj->posX -= obj->deltaPosX; 
		obj->posY -= 0x10;
		obj->velY = -10;
		CK_SetAction2(obj, &AR_ACT_SpliderFallingSlide);
	}


	RF_PlaceSprite(&obj->sde, obj->posX, obj->posY, obj->gfxChunk, false, obj->zLayer);


}

void AR_SpliderFallingSlideReact(CK_object *obj)
{
	if (obj->xDirection == motion_Right && obj->leftTI || obj->xDirection == motion_Left && obj->rightTI)
	{
		obj->velX = 0;
		CK_SetAction2(obj, &AR_ACT_SpliderAbortSlide);
	}


	if (obj->topTI)
	{
		// If it hits the ground, it immediately slows down
		obj->velY = 0;
		CK_SetAction2(obj, &AR_ACT_SpliderSlide2);
	}

	RF_PlaceSprite(&obj->sde, obj->posX, obj->posY, obj->gfxChunk, false, obj->zLayer);

}

void AR_SpliderAbortSlideReact(CK_object *obj)
{
	// Splider must transition to walking
	
	if (obj->topTI)
	{
		obj->velY = 0;
		CK_SetAction2(obj, &AR_ACT_SpliderWalk1);
		obj->timeUntillThink = US_RndT() >> 5;
	}
	

	RF_PlaceSprite(&obj->sde, obj->posX, obj->posY, obj->gfxChunk, false, obj->zLayer);

}

void AR_SpliderSlideCol(CK_object *obj1, CK_object *obj2)
{
	if (obj2->type == CT_Player)
	{
		CK_KillKeen();
	}
	else if (obj2->type == CT_Stunner)
	{
		int delX = obj2->clipRects.unitXmid - obj1->clipRects.unitXmid;

		// Splider's mouth is vulnerable during the slide
		if ((obj2->yDirection == motion_None && (obj1->xDirection == -obj2->xDirection)) ||
		 obj2->yDirection != motion_None && ((obj1->xDirection == motion_Left && delX < 0 /*&& delX > -0x100*/) || (obj1->xDirection == motion_Right && delX > 0 /*&& delX < 0x100*/))) 
		{
			CK_StunCreature(obj1, obj2, &AR_ACT_SpliderStunned);
		}

		else
		{
			CK_ShotHit(obj2);
		}
	}

}

//===========================================================================
//
// Larvbat (Skypest)
// 
//===========================================================================

CK_action AR_ACT_LarvbatFly1;
CK_action AR_ACT_LarvbatFly2;
CK_action AR_ACT_LarvbatDie1;
CK_action AR_ACT_LarvbatDie2;
CK_action AR_ACT_LarvbatDie3;
CK_action AR_ACT_LarvbatDie4;


void AR_SpawnLarvbat(int tileX, int tileY);
void AR_LarvbatFly(CK_object *obj);
void AR_LarvbatCollide(CK_object *obj1, CK_object *obj2);
void AR_LarvbatReact(CK_object *obj);

CK_action AR_ACT_LarvbatFly1 = {SPR_LARVBATL1, SPR_LARVBATR1, AT_UnscaledFrame, 0, 0, 5, 0, 0, AR_LarvbatFly, AR_LarvbatCollide, AR_LarvbatReact, &AR_ACT_LarvbatFly2};
CK_action AR_ACT_LarvbatFly2 = {SPR_LARVBATL2, SPR_LARVBATR2, AT_UnscaledFrame, 0, 0, 5, 0, 0, AR_LarvbatFly, AR_LarvbatCollide, AR_LarvbatReact, &AR_ACT_LarvbatFly1};

CK_action AR_ACT_LarvbatDie1 = {SPR_LARVBATDIE1, SPR_LARVBATDIE1, AT_UnscaledFrame, 0, 0, 8, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_LarvbatDie2};
CK_action AR_ACT_LarvbatDie2 = {SPR_LARVBATDIE2, SPR_LARVBATDIE2, AT_UnscaledFrame, 0, 0, 8, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_LarvbatDie3};
CK_action AR_ACT_LarvbatDie3 = {SPR_LARVBATDIE3, SPR_LARVBATDIE3, AT_UnscaledFrame, 0, 0, 8, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_LarvbatDie4};
CK_action AR_ACT_LarvbatDie4 = {SPR_LARVBATDIE4, SPR_LARVBATDIE4, AT_UnscaledFrame, 0, 0, 8, 0, 0, NULL, NULL, CK_BasicDrawFunc1, NULL};

void AR_SpawnLarvbat(int tileX, int tileY)
{
	CK_object *new_object = CK_GetNewObj(false);
	new_object->type = CT_Larvbat;
	new_object->active = OBJ_ACTIVE;
	new_object->zLayer = 0;
	new_object->posX = tileX << 8;
	new_object->posY = (tileY << 8);  
	new_object->xDirection = US_RndT() < 0x80 ? motion_Left : motion_Right;
	new_object->yDirection = motion_Up;
	new_object->clipped = CLIP_normal;
	CK_SetAction(new_object, &AR_ACT_LarvbatFly1);
}




void AR_LarvbatFly(CK_object *obj)
{

	if (US_RndT() < tics)
		obj->xDirection = -obj->xDirection;

	if (obj->yDirection == motion_Up && US_RndT() < tics)
		obj->yDirection = motion_Down;

	if (obj->yDirection == motion_Down && US_RndT() < tics * 2)
		obj->yDirection = -obj->yDirection;

	CK_PhysAccelHorz(obj, (int)(obj->xDirection), 20);
	CK_PhysAccelVert(obj, (int)(obj->yDirection), 20);

}

void AR_LarvbatCollide(CK_object *obj1, CK_object *obj2)
{


	if (obj2->type == CT_Stunner)
	{
		CK_SetAction2(obj1, &AR_ACT_LarvbatDie1);
		obj1->velX = obj1->velY = 0;
		CK_ShotHit(obj2);
		return;
	}
		


}

void AR_LarvbatReact(CK_object *obj)
{

	if (obj->bottomTI)
	{
		obj->velY = 8;
		obj->yDirection = motion_Down;
	}

	if (obj->leftTI)
	{
		obj->velX = 0;
		obj->xDirection = motion_Left;
	}

	if (obj->rightTI)
	{
		obj->velX = 0;
		obj->xDirection = motion_Right;
	}
	
	RF_PlaceSprite(&obj->sde, obj->posX, obj->posY, obj->gfxChunk, false, obj->zLayer);
}


//===========================================================================
//
// HORNET (wormmouth replacement)
// 
//===========================================================================
//
// Description: Walks back and forth along the ground, darts forward to attack keen when it gets close
// takes two hits to stun

void AR_SpawnHornet(int tileX, int tileY);
void AR_HornetWalk(CK_object *obj);
void AR_HornetCollide(CK_object *obj1, CK_object *obj2);
void AR_HornetReact(CK_object *obj);
void AR_DrawHitFlash(CK_object *obj, word *healthinfo, int adjX, int adjY);

CK_action AR_ACT_HornetWalk1;
CK_action AR_ACT_HornetWalk2;

CK_action AR_ACT_HornetAttack1;
CK_action AR_ACT_HornetAttack2;
CK_action AR_ACT_HornetAttack3;
CK_action AR_ACT_HornetAttack4;
CK_action AR_ACT_HornetAttack5;


CK_action AR_ACT_HornetAttackStop1;
CK_action AR_ACT_HornetAttackStop2;
CK_action AR_ACT_HornetAttackStop3;
CK_action AR_ACT_HornetAttackStop4;
CK_action AR_ACT_HornetAttackStop5;

CK_action AR_ACT_HornetStunned1;

CK_action AR_ACT_HornetWalk1 = {SPR_HORNETWALKL1, SPR_HORNETWALKR1, AT_ScaledFrame, 0, 1, 4, 12, 0, AR_HornetWalk, AR_HornetCollide, AR_HornetReact, &AR_ACT_HornetWalk2};
CK_action AR_ACT_HornetWalk2 = {SPR_HORNETWALKL2, SPR_HORNETWALKR2, AT_ScaledFrame, 0, 1, 4, 12, 0, AR_HornetWalk, AR_HornetCollide, AR_HornetReact, &AR_ACT_HornetWalk1};

CK_action AR_ACT_HornetAttack1 = {SPR_HORNETSLIDEL1, SPR_HORNETSLIDER1, AT_ScaledFrame, 0, 1, 4, 30, 0, NULL, AR_HornetCollide, AR_HornetReact, &AR_ACT_HornetAttack2};
CK_action AR_ACT_HornetAttack2 = {SPR_HORNETSLIDEL2, SPR_HORNETSLIDER2, AT_ScaledFrame, 0, 1, 4, 30, 0, NULL, AR_HornetCollide, AR_HornetReact, &AR_ACT_HornetAttack3};
CK_action AR_ACT_HornetAttack3 = {SPR_HORNETSLIDEL3, SPR_HORNETSLIDER3, AT_ScaledFrame, 0, 1, 12, 30, 0, NULL, AR_HornetCollide, AR_HornetReact, &AR_ACT_HornetAttack4};
CK_action AR_ACT_HornetAttack4 = {SPR_HORNETSLIDEL2, SPR_HORNETSLIDER2, AT_ScaledFrame, 0, 1, 4, 30, 0, NULL, AR_HornetCollide, AR_HornetReact, &AR_ACT_HornetAttack5};
CK_action AR_ACT_HornetAttack5 = {SPR_HORNETSLIDEL1, SPR_HORNETSLIDER1, AT_ScaledFrame, 0, 1, 4, 30, 0, NULL, AR_HornetCollide, AR_HornetReact, &AR_ACT_HornetWalk1};

CK_action AR_ACT_HornetAttackStop1 = {SPR_HORNETSLIDER1, SPR_HORNETSLIDEL1, AT_ScaledFrame, 0, 1, 4, 0, 0, NULL, AR_HornetCollide, AR_HornetReact, &AR_ACT_HornetAttackStop2};
CK_action AR_ACT_HornetAttackStop2 = {SPR_HORNETSLIDER2, SPR_HORNETSLIDEL2, AT_ScaledFrame, 0, 1, 4, 0, 0, NULL, AR_HornetCollide, AR_HornetReact, &AR_ACT_HornetAttackStop3};
CK_action AR_ACT_HornetAttackStop3 = {SPR_HORNETSLIDER3, SPR_HORNETSLIDEL3, AT_ScaledFrame, 0, 1, 12, 0, 0, NULL, AR_HornetCollide, AR_HornetReact, &AR_ACT_HornetAttackStop4};
CK_action AR_ACT_HornetAttackStop4 = {SPR_HORNETSLIDER2, SPR_HORNETSLIDEL2, AT_ScaledFrame, 0, 1, 4, 0, 0, NULL, AR_HornetCollide, AR_HornetReact, &AR_ACT_HornetAttackStop5};
CK_action AR_ACT_HornetAttackStop5 = {SPR_HORNETSLIDER1, SPR_HORNETSLIDEL1, AT_ScaledFrame, 0, 1, 4, 0, 0, NULL, AR_HornetCollide, AR_HornetReact, &AR_ACT_HornetWalk1};

CK_action AR_ACT_HornetStunned1 = {SPR_HORNETDEAD, SPR_HORNETDEAD, AT_Frame, 0, 0, 0, 0, 0, CK_Fall, NULL, CK_BasicDrawFunc4, NULL};

// user1 = health



void AR_SpawnHornet(int tileX, int tileY)
{
	CK_object *new_object = CK_GetNewObj(false);
	new_object->type = CT_Hornet;
	new_object->active = OBJ_ACTIVE;
	new_object->zLayer = 0;
	new_object->posX = tileX << 8;
	new_object->posY = (tileY << 8) - 0x100; 
	new_object->xDirection = US_RndT() < 0x80 ? motion_Left : motion_Right;
	new_object->yDirection = motion_None;
	new_object->clipped = CLIP_normal;
	new_object->user1 = 2; 
	CK_SetAction(new_object, &AR_ACT_HornetWalk1);
}

void AR_HornetWalk(CK_object *obj)
{
	unsigned dx;
	int dy;

	// Slide at Keen if he gets within range
	
	dx = ck_keenObj->clipRects.unitXmid - obj->clipRects.unitXmid;
	dy = ck_keenObj->clipRects.unitY2 - obj->clipRects.unitY2;


	if (ABS(dy) < 0x200 && (obj->xDirection == motion_Right && dx < 0x400 || obj->xDirection == motion_Left && dx > 0xFC00))
	{
		//Play a sound
		obj->timeUntillThink = 30 + 20;
		SD_PlaySound(HORNETATTACKSND);
		CK_SetAction2(obj, &AR_ACT_HornetAttack1);
		return;
	}


}

void AR_HornetCollide(CK_object *obj1, CK_object *obj2)
{
	if (obj2->type == CT_Player)
	{
		// Fudgy check to see if contact was not made by Keen hanging on to
		// ledge 

		if ((obj2->currentAction == &CK_ACT_keenHang1 || obj2->currentAction == &CK_ACT_keenHang2)
		 && (obj2->posY > obj1->posY + 0x100))
		{
			return;
		}

		CK_KillKeen();
		return;
	}

	if (obj2->type == CT_Stunner)
	{
		int flashtime, health;

		health = obj1->user1 & 0xFF;

		if (--health == 0)
		{
			CK_StunCreature(obj1, obj2, &AR_ACT_HornetStunned1);
			return;
		}
		else
		{
			obj1->user1 = (10 << 8) + health;
			CK_ShotHit(obj2);
			return;
		}
	}
}


void AR_HornetReact(CK_object *obj)
{
	// Hit wall walking right; turn around and go left
	if (obj->xDirection == motion_Right && obj->leftTI != 0)
	{
		obj->posX -= obj->deltaPosX;
		if (obj->currentAction >= &AR_ACT_HornetAttack1 && obj->currentAction <= &AR_ACT_HornetAttack5)
			obj->xDirection = motion_Right;
		else
		{
		obj->xDirection = motion_Left;
		obj->timeUntillThink = US_RndT() / 32;
		CK_SetAction2(obj, obj->currentAction);
		}
	}
		// Hit wall walking left; turn around and go right
	else if (obj->xDirection == motion_Left && obj->rightTI != 0)
	{
		obj->posX -= obj->deltaPosX;
		if (obj->currentAction >= &AR_ACT_HornetAttack1 && obj->currentAction <= &AR_ACT_HornetAttack5)
			obj->xDirection = motion_Left;
		else
		{
		obj->xDirection = motion_Right;
		obj->timeUntillThink = US_RndT() / 32;
		CK_SetAction2(obj, obj->currentAction);
		}
	}
		// Walked off of ledge; turn around
	else if (obj->topTI == 0)
	{
		obj->posX -= obj->deltaPosX;
		obj->xDirection = -obj->xDirection;
		obj->timeUntillThink = US_RndT() / 32;
		if (obj->currentAction >= &AR_ACT_HornetAttack1 && obj->currentAction <= &AR_ACT_HornetAttack5)
			CK_SetAction2(obj, &AR_ACT_HornetAttackStop1 + ((int)obj->currentAction - (int)&AR_ACT_HornetAttack1)/sizeof(CK_action)); 
		else
			CK_SetAction2(obj, obj->currentAction);
	}
	
	AR_DrawHitFlash(obj, (word *)&(obj->user1), 0, -0x80);

}

void AR_DrawHitFlash(CK_object *obj, word *healthinfo, int adjX, int adjY)
{
	// high byte of healthinfo holds remaining whileflash time; lowbyte holds health

	int flashtime = *healthinfo >> 8;

	if ((flashtime -= tics) < 0)
		flashtime = 0;

	*healthinfo = (*healthinfo & 0xFF) | (flashtime << 8);

	RF_PlaceSprite(&(obj->sde), obj->posX + adjX, obj->posY + adjY, obj->gfxChunk, flashtime?whitedraw:spritedraw, obj->zLayer);


}


//===========================================================================
//
// DARTGUNS
// 
//===========================================================================
CK_action AR_ACT_DartHorz1;
CK_action AR_ACT_DartHorz2;

CK_action AR_ACT_DartUp1;
CK_action AR_ACT_DartUp2;

CK_action AR_ACT_DartDown1;
CK_action AR_ACT_DartDown2;
CK_action AR_ACT_DartGun;

void AR_DartGun(CK_object *obj);

CK_action AR_ACT_DartGun = {NULL, NULL, AT_ScaledOnce, 0, 0, 150, 0, 0, AR_DartGun, NULL, NULL, &AR_ACT_DartGun};

CK_action AR_ACT_DartHorz1 = {DARTEAST1SPR, DARTWEST1SPR, AT_ScaledOnce, true, 0, 6, 64, 0, NULL, CK_DeadlyCol2, CK_RemoveOnContact, &AR_ACT_DartHorz2};
CK_action AR_ACT_DartHorz2 = {DARTEAST2SPR, DARTWEST2SPR, AT_ScaledOnce, true, 0, 6, 64, 0, NULL, CK_DeadlyCol2, CK_RemoveOnContact, &AR_ACT_DartHorz1};

CK_action AR_ACT_DartUp1 = {DARTNORTH1SPR, DARTNORTH1SPR, AT_ScaledOnce, true, 0, 6, 0, 64, NULL, CK_DeadlyCol2, CK_RemoveOnContact, &AR_ACT_DartUp2};
CK_action AR_ACT_DartUp2 = {DARTNORTH2SPR, DARTNORTH2SPR, AT_ScaledOnce, true, 0, 6, 0, 64, NULL, CK_DeadlyCol2, CK_RemoveOnContact, &AR_ACT_DartUp1};

CK_action AR_ACT_DartDown1 = {DARTSOUTH1SPR, DARTSOUTH1SPR, AT_ScaledOnce, true, 0, 6, 0, 64, NULL, CK_DeadlyCol2, CK_RemoveOnContact, &AR_ACT_DartDown2};
CK_action AR_ACT_DartDown2 = {DARTSOUTH2SPR, DARTSOUTH2SPR, AT_ScaledOnce, true, 0, 6, 0, 64, NULL, CK_DeadlyCol2, CK_RemoveOnContact, &AR_ACT_DartDown1};

void AR_SpawnDartgun (int tileX, int tileY, int dir)
{
	CK_object *newobj;

	newobj = CK_GetNewObj(false);
	newobj->type = CT_Dartgun;
	newobj->active = OBJ_ACTIVE;
	newobj->clipRects.tileX1 = newobj->clipRects.tileX2 = tileX;
	newobj->clipRects.tileY1 = newobj->clipRects.tileY2 = tileY;
	newobj->posX = newobj->clipRects.unitX1 = newobj->clipRects.unitX2 = tileX << G_T_SHIFT;
	newobj->posY = newobj->clipRects.unitY1 = newobj->clipRects.unitY2 = tileY << G_T_SHIFT;
	newobj->user1 = dir;
	CK_SetAction(newobj, &AR_ACT_DartGun);
}


void AR_DartGun(CK_object *obj)
{

	CK_object *newobj;

	newobj = CK_GetNewObj(true);
	newobj->type = CT_EnemyShot;
	newobj->active = OBJ_EXISTS_ONLY_ONSCREEN;
	newobj->posX = obj->posX;
	newobj->posY = obj->posY;
	newobj->clipped = CLIP_dart;

	switch (obj->user1)
	{
	case 0:
		newobj->xDirection = motion_None;
		newobj->yDirection = motion_Up;
		newobj->posX += 0x80;
		CK_SetAction(newobj, &AR_ACT_DartUp1);
		break;
	case 1:
		newobj->xDirection = motion_Right;
		newobj->yDirection = motion_None;
		newobj->posY += 0x40;
		CK_SetAction(newobj, &AR_ACT_DartHorz1);
		break;
	case 2:
		newobj->xDirection = motion_None;
		newobj->yDirection = motion_Down;
		newobj->posX += 0x80;
		CK_SetAction(newobj, &AR_ACT_DartDown1);
		break;
	case 3:
		newobj->xDirection = motion_Left;
		newobj->yDirection = motion_None;
		newobj->posY += 0x40;
		CK_SetAction(newobj, &AR_ACT_DartHorz1);
		break;
	}

	SD_PlaySound(DARTSHOOTSND);
}


