/* Keen Dreams Source Code
 * Copyright (C) 2014 Javier M. Chavez
 * Copyright (C) 1993-2014 Flat Rock Software
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

// KD_MISC.C

#include "KD_DEF.H"
#pragma	hdrstop

/*
=============================================================================

LOCAL CONSTANTS

=============================================================================
*/


/*
=============================================================================

GLOBAL VARIABLES

=============================================================================
*/


/*
=============================================================================

LOCAL VARIABLES

=============================================================================
*/


//===========================================================================
//
// Atroxian (Bluebird replacement)
// 
//===========================================================================

// Description: Files around like a keen 4 bluebird, seeking keen.
// If it's hit by a stunner, it's temporarily stunned and it falls to the ground.
// While stunned, if hit by 5 hits, then transforms to bipedal state
// If hit by 5 hits in bipedal state, then it is stunned for good

// user1 = flying variable
// user2 = packed health format (see Hornet)

void AR_SpawnAtroxian(int tileX, int tileY);
void AR_AtroxianFly(CK_object *obj);
void AR_AtroxianRecoverToFlying (CK_object *obj);
void AR_AtroxianExploding (CK_object *obj);
void AR_AtroxianWalk(CK_object *obj);
void AR_AtroxianFlyingCollide(CK_object *obj1, CK_object *obj2);
void AR_AtroxianStunnedCollide(CK_object *obj1, CK_object *obj2);
void AR_AtroxianWalkingCollide(CK_object *obj1, CK_object *obj2);
void AR_AtroxianFlyingReact(CK_object *obj);
void AR_AtroxianStunnedReact(CK_object *obj);
void AR_AtroxianBasicReact (CK_object *obj);
void AR_AtroxianExplodingReact (CK_object *obj);
void AR_AtroxianPreWalkReact(CK_object *obj);
void AR_AtroxianWalkingReact(CK_object *obj);

CK_action AR_ACT_AtroxianFly1;
CK_action AR_ACT_AtroxianFly2;
CK_action AR_ACT_AtroxianFly3;
CK_action AR_ACT_AtroxianFly4;

CK_action AR_ACT_AtroxianStunnedFall1;

CK_action AR_ACT_AtroxianStunned1;
CK_action AR_ACT_AtroxianRecover1;
CK_action AR_ACT_AtroxianRecover2;
CK_action AR_ACT_AtroxianRecover3;
CK_action AR_ACT_AtroxianRecover4;

CK_action AR_ACT_AtroxianWalkPre1;
CK_action AR_ACT_AtroxianWalkPre2;

CK_action AR_ACT_AtroxianWalk1;
CK_action AR_ACT_AtroxianWalk2;
CK_action AR_ACT_AtroxianWalk3;
CK_action AR_ACT_AtroxianWalk4;

CK_action AR_ACT_AtroxianDead1;

CK_action AR_ACT_SmallExplosion1;
CK_action AR_ACT_SmallExplosion2;
CK_action AR_ACT_SmallExplosion3;

// K1n9_Duk3 fix: don't need these anymore
/*
CK_action AR_ACT_AtroxianBigExplosion1;
CK_action AR_ACT_AtroxianBigExplosion2;
CK_action AR_ACT_AtroxianBigExplosion3;
CK_action AR_ACT_AtroxianBigExplosion4;
CK_action AR_ACT_AtroxianBigExplosion5;
*/


//Flying
CK_action AR_ACT_AtroxianFly1 = {SPR_ATROXIANFLY1, SPR_ATROXIANFLY1, AT_ScaledFrame, 0, 0, 8, 0, 0, AR_AtroxianFly, AR_AtroxianFlyingCollide, AR_AtroxianFlyingReact, &AR_ACT_AtroxianFly2};
CK_action AR_ACT_AtroxianFly2 = {SPR_ATROXIANFLY2, SPR_ATROXIANFLY2, AT_ScaledFrame, 0, 0, 8, 0, 0, AR_AtroxianFly, AR_AtroxianFlyingCollide, AR_AtroxianFlyingReact, &AR_ACT_AtroxianFly3};
CK_action AR_ACT_AtroxianFly3 = {SPR_ATROXIANFLY3, SPR_ATROXIANFLY3, AT_ScaledFrame, 0, 0, 8, 0, 0, AR_AtroxianFly, AR_AtroxianFlyingCollide, AR_AtroxianFlyingReact, &AR_ACT_AtroxianFly4};
CK_action AR_ACT_AtroxianFly4 = {SPR_ATROXIANFLY4, SPR_ATROXIANFLY4, AT_ScaledFrame, 0, 0, 8, 0, 0, AR_AtroxianFly, AR_AtroxianFlyingCollide, AR_AtroxianFlyingReact, &AR_ACT_AtroxianFly1};

// Stunned (temporarily)
CK_action AR_ACT_AtroxianStunnedFall1 = {SPR_ATROXIANSTUNNED, SPR_ATROXIANSTUNNED, AT_Frame, 0, 0, 8, 0, 0, CK_Fall2, AR_AtroxianStunnedCollide, AR_AtroxianStunnedReact, &AR_ACT_AtroxianStunnedFall1};

CK_action AR_ACT_AtroxianStunned1 = {SPR_ATROXIANSTUNNED, SPR_ATROXIANSTUNNED, AT_UnscaledFrame, 0, 0, 350, 0, 0, CK_Fall2, AR_AtroxianStunnedCollide, AR_AtroxianBasicReact, &AR_ACT_AtroxianRecover1};

CK_action AR_ACT_AtroxianRecover1 = {SPR_ATROXIANFLY1, SPR_ATROXIANFLY1, AT_UnscaledOnce, 0, 0, 16, 0, 0, NULL, AR_AtroxianStunnedCollide, AR_AtroxianBasicReact, &AR_ACT_AtroxianRecover2};
CK_action AR_ACT_AtroxianRecover2 = {SPR_ATROXIANSTUNNED, SPR_ATROXIANSTUNNED, AT_UnscaledOnce, 0, 0, 16, 0, 0, NULL, AR_AtroxianStunnedCollide, AR_AtroxianBasicReact, &AR_ACT_AtroxianRecover3};
CK_action AR_ACT_AtroxianRecover3 = {SPR_ATROXIANFLY1, SPR_ATROXIANFLY1, AT_UnscaledOnce, 0, 0, 16, 0, 0, NULL, AR_AtroxianStunnedCollide, AR_AtroxianBasicReact, &AR_ACT_AtroxianRecover4};
CK_action AR_ACT_AtroxianRecover4 = {SPR_ATROXIANSTUNNED, SPR_ATROXIANSTUNNED, AT_UnscaledOnce, 0, 0, 16, 0, 0, NULL, AR_AtroxianStunnedCollide, AR_AtroxianBasicReact, &AR_ACT_AtroxianFly1};


// Transition to Walking (when hit ground)
CK_action AR_ACT_AtroxianWalkPre1 = {SPR_ATROXIANSTUNNED, SPR_ATROXIANSTUNNED, AT_ScaledFrame, 0, 0, 70, 0, 0, AR_AtroxianExploding, NULL, AR_AtroxianExplodingReact, &AR_ACT_AtroxianWalkPre2 };
CK_action AR_ACT_AtroxianWalkPre2 = {SPR_ATROXIANSTUNNED, SPR_ATROXIANSTUNNED, AT_Frame, 0, 0, 8, 0, 0, CK_Fall2, AR_AtroxianWalkingCollide, AR_AtroxianPreWalkReact, &AR_ACT_AtroxianStunnedFall1};

// Walking state
CK_action AR_ACT_AtroxianWalk1 = {SPR_ATROXIANWALKL1, SPR_ATROXIANWALKR1, AT_ScaledOnce, 0, 1, 10, 16, 0, AR_AtroxianWalk, AR_AtroxianWalkingCollide, AR_AtroxianWalkingReact, &AR_ACT_AtroxianWalk2};
CK_action AR_ACT_AtroxianWalk2 = {SPR_ATROXIANWALKL2, SPR_ATROXIANWALKR2, AT_ScaledOnce, 0, 1, 10, 16, 0, AR_AtroxianWalk, AR_AtroxianWalkingCollide, AR_AtroxianWalkingReact, &AR_ACT_AtroxianWalk3};
CK_action AR_ACT_AtroxianWalk3 = {SPR_ATROXIANWALKL3, SPR_ATROXIANWALKR3, AT_ScaledOnce, 0, 1, 10, 16, 0, AR_AtroxianWalk, AR_AtroxianWalkingCollide, AR_AtroxianWalkingReact, &AR_ACT_AtroxianWalk4};
CK_action AR_ACT_AtroxianWalk4 = {SPR_ATROXIANWALKL4, SPR_ATROXIANWALKR4, AT_ScaledOnce, 0, 1, 10, 16, 0, AR_AtroxianWalk, AR_AtroxianWalkingCollide, AR_AtroxianWalkingReact, &AR_ACT_AtroxianWalk1};

// Dead
CK_action AR_ACT_AtroxianDead1 = {SPR_ATROXIANDEAD, SPR_ATROXIANDEAD, AT_Frame, 0, 0, 0, 0, 0, CK_Fall, NULL, CK_BasicDrawFunc1, NULL};


// Explosion Sprites
CK_action AR_ACT_SmallExplosion1 = {SMALLEXPLOSION1SPR, SMALLEXPLOSION1SPR, AT_UnscaledOnce, 0, 0, 4, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_SmallExplosion2};
CK_action AR_ACT_SmallExplosion2 = {SMALLEXPLOSION2SPR, SMALLEXPLOSION2SPR, AT_UnscaledOnce, 0, 0, 4, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_SmallExplosion3};
CK_action AR_ACT_SmallExplosion3 = {SMALLEXPLOSION3SPR, SMALLEXPLOSION3SPR, AT_UnscaledOnce, 0, 0, 4, 0, 0, NULL, NULL, CK_BasicDrawFunc1, NULL};

// K1n9_Duk3 fix: don't need these anymore
/*
CK_action AR_ACT_AtroxianBigExplosion1 = {SPR_MAGMINEEXPL1, SPR_MAGMINEEXPL1, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_AtroxianBigExplosion2};
CK_action AR_ACT_AtroxianBigExplosion2 = {SPR_MAGMINEEXPL2, SPR_MAGMINEEXPL2, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_AtroxianBigExplosion3};
CK_action AR_ACT_AtroxianBigExplosion3 = {SPR_MAGMINEEXPL3, SPR_MAGMINEEXPL3, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_AtroxianBigExplosion4};
CK_action AR_ACT_AtroxianBigExplosion4 = {SPR_MAGMINEEXPL4, SPR_MAGMINEEXPL4, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_AtroxianBigExplosion5};
CK_action AR_ACT_AtroxianBigExplosion5 = {SPR_MAGMINEEXPL5, SPR_MAGMINEEXPL5, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, NULL, CK_BasicDrawFunc1, NULL};
*/

void AR_SpawnAtroxian(int tileX, int tileY)
{
	CK_object *new_object = CK_GetNewObj(false);
	new_object->type = CT_Atroxian;
	new_object->active = OBJ_ACTIVE;
	new_object->zLayer = 0;
	new_object->posX = tileX << 8;
	new_object->posY = (tileY << 8);  
	new_object->xDirection = US_RndT() < 0x80 ? motion_Left : motion_Right;
	new_object->yDirection = motion_Up;
	//new_object->clipped = CLIP_simple;
	new_object->user1 = 0;
	new_object->user2 = 5;
	CK_SetAction(new_object, &AR_ACT_AtroxianFly1);
}

void AR_AtroxianFly(CK_object *obj)
{
	if (obj->user1)
	{
		if (obj->clipRects.unitXmid < ck_keenObj->clipRects.unitXmid)
			obj->xDirection = motion_Right;
		else
			obj->xDirection = motion_Left;
	}

	CK_PhysAccelHorz(obj, obj->xDirection, 16);

	if (obj->posY < ck_keenObj->posY - 0x100)
		CK_PhysAccelVert(obj, motion_Down, 16);
	else
		CK_PhysAccelVert(obj, motion_Up, 16);

}


void AR_AtroxianExploding(CK_object *obj)
{
	CK_Fall(obj);

	if ((obj->user4 -= tics) == 0)
	{
		CK_object *newobj = CK_GetNewObj(true);
		obj->user4 = 20;
		newobj->posX = obj->clipRects.unitX1 - 0x80 + ((obj->clipRects.unitX2 - obj->clipRects.unitX1) * US_RndT() >> 8);
		newobj->posY = obj->clipRects.unitY1 - 0x100 + ((obj->clipRects.unitY2 - obj->clipRects.unitY1) * US_RndT() >> 8);
		newobj->type = CT_Friendly;
		newobj->clipped = CLIP_not;
		newobj->active = OBJ_EXISTS_ONLY_ONSCREEN;
		newobj->zLayer = 2;
		CK_SetAction(newobj, &AR_ACT_SmallExplosion1);
		SD_PlaySound(ATROXIANGROUNDEDSND);
	}
}

void AR_AtroxianRecoverToFlying (CK_object *obj)
{
	obj->velY = -20;
	CK_SetAction2(obj, &AR_ACT_AtroxianFly1);
	ck_nextX = 0;
	ck_nextY = -0x80;
	CK_PhysFullClipToWalls(obj);
}


void AR_AtroxianWalk(CK_object *obj)
{
	if (US_RndT() < 0x20)
	{
		
		// Follow keen
		int delX = ck_keenObj->clipRects.unitXmid - obj->clipRects.unitXmid;
		// Follow Keen
		if (delX > 0x200)
			obj->xDirection = motion_Right;
		else if (delX < -0x200)
			obj->xDirection = motion_Left;
	}

}

void AR_AtroxianFlyingCollide(CK_object *obj1, CK_object *obj2)
{
	if (obj2->type == CT_Player)
	{
		CK_KillKeen();
		return;
	}

	if (obj2->type == CT_Stunner)
	{
		CK_ShotHit(obj2);
		CK_SetAction2(obj1, &AR_ACT_AtroxianStunnedFall1);
		obj1->velX = 0;
		obj1->velY = -20;
	}


}

void AR_AtroxianStunnedCollide(CK_object *obj1, CK_object *obj2)
{
	int health;

	if (obj2->type == CT_Stunner)
	{
		CK_ShotHit(obj2);

		health = obj1->user2 & 0xFF;

		if (--health == 0)
		{
			// Set new health to 5
			obj1->user2 = 5;
			obj1->clipped = CLIP_normal;
			obj1->user4 = 10;
			CK_SetAction2(obj1, &AR_ACT_AtroxianWalkPre1);
			// TODO: Spawn an explosion sprite here
		}
		else
		{
			obj1->user2 = (14 << 8) | health; 
		}
	}

}


void AR_AtroxianWalkingCollide(CK_object *obj1, CK_object *obj2)
{
	extern CK_action AR_ACT_MagmineExplosion1;	// K1n9_Duk3 fix

	int health;
	if (obj2->type == CT_Player)
	{
		CK_KillKeen();
		return;
	}


	if (obj2->type == CT_Stunner)
	{
		CK_ShotHit(obj2);

		health = obj1->user2 & 0xFF;

		if (--health == 0)
		{
			CK_object *newobj = CK_GetNewObj(true);

			newobj->type = CT_Friendly;
			newobj->active = OBJ_EXISTS_ONLY_ONSCREEN;
			newobj->clipped = CLIP_not;
			newobj->posX = obj1->posX + 0xC0;
			newobj->posY = obj1->posY;
			newobj->zLayer = 2;
			// K1n9_Duk3 fix (requested by Gridlock):
			//CK_SetAction(newobj, &AR_ACT_AtroxianBigExplosion1);
			CK_SetAction(newobj, &AR_ACT_MagmineExplosion1);

			CK_SetAction(obj1, &AR_ACT_AtroxianDead1);
			SD_PlaySound(MAGMINEEXPLODESND);
		}
		else
		{
			obj1->user2 = (14 << 8) | health; 
		}
		return;

	}

}


void AR_AtroxianFlyingReact(CK_object *obj)
{

	if ((obj->bottomTI || obj->topTI) && !obj->user1)
		obj->user1++;


	if (obj->rightTI && obj->velX < 0 || obj->leftTI && obj->velX > 0)
	{
		obj->velX = 0;
		obj->xDirection = -obj->xDirection;
	}

	if (!obj->bottomTI && !obj->topTI)
		obj->user1 = 0;

	AR_DrawHitFlash(obj, (word *)&(obj->user2), 0, 0);
	//RF_PlaceSprite(&obj->sde, obj->posX, obj->posY, obj->gfxChunk, false, obj->zLayer);

}


void AR_AtroxianStunnedReact(CK_object *obj)
{
	if (obj->topTI)
	{
		CK_SetAction2(obj, &AR_ACT_AtroxianStunned1);
	}

	AR_DrawHitFlash(obj, (word *)&(obj->user2), 0, 0);
	//RF_PlaceSprite(&obj->sde, obj->posX, obj->posY, obj->gfxChunk, false, obj->zLayer);
}

void AR_AtroxianBasicReact (CK_object *obj)
{
	AR_DrawHitFlash(obj, (word *)&(obj->user2), 0, 0);
}

void AR_AtroxianExplodingReact(CK_object *obj)
{
	RF_PlaceSprite(&obj->sde, obj->posX, obj->posY, obj->gfxChunk, false, obj->zLayer);
	

}

void AR_AtroxianPreWalkReact(CK_object *obj)
{
	if (obj->topTI)
		CK_SetAction2(obj, &AR_ACT_AtroxianWalk1);

	RF_PlaceSprite(&obj->sde, obj->posX, obj->posY, obj->gfxChunk, false, obj->zLayer);

}

void AR_AtroxianWalkingReact(CK_object *obj)
{

	// Hit wall walking right; turn around and go left
	if (obj->xDirection == motion_Right && obj->leftTI != 0)
	{
		obj->posX -= obj->deltaPosX;
		obj->xDirection = motion_Left;
		obj->timeUntillThink = US_RndT() / 32;
		CK_SetAction2(obj, obj->currentAction);
	}
		// Hit wall walking left; turn around and go right
	else if (obj->xDirection == motion_Left && obj->rightTI != 0)
	{
		obj->posX -= obj->deltaPosX;
		obj->xDirection = motion_Right;
		obj->timeUntillThink = US_RndT() / 32;
		CK_SetAction2(obj, obj->currentAction);
	}
		// Walked off of ledge; turn around
	else if (obj->topTI == 0)
	{
		obj->posX -= obj->deltaPosX;
		obj->xDirection = -obj->xDirection;
		obj->timeUntillThink = US_RndT() / 32;
		CK_SetAction2(obj, obj->currentAction);
	}
	
	AR_DrawHitFlash(obj, (word *)&(obj->user2), 0x80, 0x40);

}


//===========================================================================
//
// Atroxite (Lick replacement)
// 
//===========================================================================

// Description:  Jumps in a high arc; when it gets close to keen, it slides forward along the ground
CK_action AR_ACT_AtroxiteHop1;
CK_action AR_ACT_AtroxiteHop2;
CK_action AR_ACT_AtroxiteHop3;


CK_action AR_ACT_AtroxiteAttack1;
CK_action AR_ACT_AtroxiteAttack2;
CK_action AR_ACT_AtroxiteAttack3;
CK_action AR_ACT_AtroxiteAttack4;
CK_action AR_ACT_AtroxiteAttack5;
CK_action AR_ACT_AtroxiteAttack6;
CK_action AR_ACT_AtroxiteAttack7;

CK_action AR_ACT_AtroxiteStunned1;


void AR_SpawnAtroxite(int tileX, int tileY);
void AR_AtroxiteMove(CK_object *obj);
void AR_AtroxiteAttack(CK_object *obj);
void AR_AtroxiteCollide(CK_object *obj1, CK_object *obj2);
void AR_AtroxiteJumpReact(CK_object *obj1);


CK_action AR_ACT_AtroxiteHop1 = {SPR_ATROXITEJUMPL1, SPR_ATROXITEJUMPR1, AT_UnscaledOnce, 0, 0, 10, 0, 0, AR_AtroxiteMove, AR_AtroxiteCollide, CK_BasicDrawFunc1, &AR_ACT_AtroxiteHop2};
CK_action AR_ACT_AtroxiteHop2 = {SPR_ATROXITEJUMPL2, SPR_ATROXITEJUMPR2, AT_Frame, 0, 0, 10, 0, 0, CK_Fall, AR_AtroxiteCollide, AR_AtroxiteJumpReact, &AR_ACT_AtroxiteHop3};
CK_action AR_ACT_AtroxiteHop3 = {SPR_ATROXITEJUMPL4, SPR_ATROXITEJUMPR4, AT_UnscaledOnce, 0, 0, 10, 0, 0, NULL, AR_AtroxiteCollide, CK_BasicDrawFunc1, &AR_ACT_AtroxiteHop1};


CK_action AR_ACT_AtroxiteAttack1 = {SPR_ATROXITEATTACKL2, SPR_ATROXITEATTACKR2, AT_ScaledFrame, 0, 1, 4, 20, 0, AR_AtroxiteAttack, AR_AtroxiteCollide, CK_BasicDrawFunc1, &AR_ACT_AtroxiteAttack2};
CK_action AR_ACT_AtroxiteAttack2 = {SPR_ATROXITEATTACKL3, SPR_ATROXITEATTACKR3, AT_ScaledFrame, 0, 1, 4, 20, 0, AR_AtroxiteAttack, AR_AtroxiteCollide, CK_BasicDrawFunc1, &AR_ACT_AtroxiteAttack3};
CK_action AR_ACT_AtroxiteAttack3 = {SPR_ATROXITEATTACKL2, SPR_ATROXITEATTACKR2, AT_ScaledFrame, 0, 1, 4, 20, 0, AR_AtroxiteAttack, AR_AtroxiteCollide, CK_BasicDrawFunc1, &AR_ACT_AtroxiteAttack4};
CK_action AR_ACT_AtroxiteAttack4 = {SPR_ATROXITEATTACKL1, SPR_ATROXITEATTACKR1, AT_ScaledFrame, 0, 1, 4, 20, 0, AR_AtroxiteAttack, AR_AtroxiteCollide, CK_BasicDrawFunc1, &AR_ACT_AtroxiteAttack5};
CK_action AR_ACT_AtroxiteAttack5 = {SPR_ATROXITEATTACKL2, SPR_ATROXITEATTACKR2, AT_ScaledFrame, 0, 1, 4, 20, 0, AR_AtroxiteAttack, AR_AtroxiteCollide, CK_BasicDrawFunc1, &AR_ACT_AtroxiteAttack6};
CK_action AR_ACT_AtroxiteAttack6 = {SPR_ATROXITEATTACKL3, SPR_ATROXITEATTACKR3, AT_ScaledFrame, 0, 1, 4, 20, 0, AR_AtroxiteAttack, AR_AtroxiteCollide, CK_BasicDrawFunc1, &AR_ACT_AtroxiteAttack7};
CK_action AR_ACT_AtroxiteAttack7 = {SPR_ATROXITEATTACKL2, SPR_ATROXITEATTACKR2, AT_ScaledFrame, 0, 1, 4, 20, 0, AR_AtroxiteAttack, AR_AtroxiteCollide, CK_BasicDrawFunc1, &AR_ACT_AtroxiteHop3};

// Stunned
CK_action AR_ACT_AtroxiteStunned1 = {SPR_ATROXITEDEAD, SPR_ATROXITEDEAD, AT_Frame, 0, 0, 0, 0, 0, CK_Fall, NULL, CK_BasicDrawFunc4, NULL};

void AR_SpawnAtroxite(int tileX, int tileY)
{
	CK_object *new_object = CK_GetNewObj(false);
	new_object->type = CT_Atroxite;
	new_object->active = OBJ_ACTIVE;
	new_object->zLayer = 0;
	new_object->posX = tileX << 8;
	new_object->posY = (tileY << 8);  
	new_object->xDirection = US_RndT() < 0x80 ? motion_Left : motion_Right;
	new_object->yDirection = motion_None;
	new_object->clipped = CLIP_normal;
	CK_SetAction(new_object, &AR_ACT_AtroxiteHop2);
}


void AR_AtroxiteMove(CK_object *obj)
{
	int dx, dy;

	dx = ck_keenObj->clipRects.unitXmid - obj->clipRects.unitXmid;	
	dy = ck_keenObj->clipRects.unitY1 - obj->clipRects.unitY1;

	// follow keen
	if (dx > 0)
		obj->xDirection = motion_Right;
	else
		obj->xDirection = motion_Left;


	// Attack if within range
	if (dy >= -0x100 && dy <= 0x100 && dx >= -0x400 && dx <= 0x400)
	{
		SD_PlaySound(ATROXITEATTACKSND);
		CK_SetAction2(obj, &AR_ACT_AtroxiteAttack1);
		return;
	}


	// Otherwise, jump
	// Adapt jump height based on tiles in front

	{
	unsigned far *tile = mapsegs[1] + mapbwidthtable[obj->clipRects.tileY2]/2 + obj->clipRects.tileXmid + obj->xDirection;
	unsigned walldir = obj->xDirection == motion_Right?WESTWALL:EASTWALL;
	if (!tinf[walldir + *tile] && !tinf[walldir + *(tile - mapwidth)] && (tinf[walldir + *(tile - mapwidth*2)] || tinf[walldir + *(tile - mapwidth*3)]))
	{
		obj->velX = obj->xDirection << 4;
		obj->velY = -30;
		return;
	}
	}

	obj->velX = obj->xDirection << 3;
	obj->velY = -64;

}

void AR_AtroxiteAttack(CK_object *obj)
{

	if (ck_gameState.difficulty == 1)
		ck_nextX -= 16 * obj->xDirection;
	else if (ck_gameState.difficulty == 3)
		ck_nextX += 16 * obj->xDirection;
}

void AR_AtroxiteCollide(CK_object *obj1, CK_object *obj2)
{
	if (obj2->type == CT_Stunner)
	{
		CK_StunCreature(obj1, obj2, &AR_ACT_AtroxiteStunned1);
		SD_PlaySound(ATROXITESTUNNEDSND);
		return;
	}

	if (obj2->type == CT_Player)
	{
		CK_KillKeen();
		return;
	}

}

void AR_AtroxiteJumpReact(CK_object *obj)
{
	if (obj->topTI)
	{
		CK_SetAction2(obj, &AR_ACT_AtroxiteHop3);
		ck_nextX = ck_nextY = 0;
		obj->velX = obj->velY = 0;
	}

	else if (obj->bottomTI)
	{
		obj->velY = 0;
		obj->yDirection = motion_Down;
	}

	else if (obj->leftTI || obj->rightTI)
	{
		// Just keep moving over walls

		//obj->xDirection = -obj->xDirection;
		//obj->velX = -obj->velX;
	}

	RF_PlaceSprite(&obj->sde, obj->posX, obj->posY, obj->gfxChunk, false, obj->zLayer);

}


//===========================================================================
//
// Axoclaw (Thundercloud replacement)
// 
//===========================================================================


// Description: Hovers in the air, moves back and forth.
// Periodically shoots a homing missile that moves at a constant horizontal speed and homes in on keen vertically

void AR_SpawnAxoclaw(int tileX, int tileY);
void AR_AxoclawThink(CK_object *obj);
void AR_AxoclawShoot(CK_object *obj);
void AR_AxoclawReact(CK_object *obj);
void AR_AxoclawShot(CK_object *obj);
void AR_AxoclawShotReact(CK_object *obj);


CK_action AR_ACT_AxoclawMove1;
CK_action AR_ACT_AxoclawMove2;


CK_action AR_ACT_AxoclawShoot1;
CK_action AR_ACT_AxoclawShoot2;
CK_action AR_ACT_AxoclawShoot3;
CK_action AR_ACT_AxoclawShoot4;
CK_action AR_ACT_AxoclawShoot5;
CK_action AR_ACT_AxoclawShoot6;

CK_action AR_ACT_AxoclawShotMove1;
CK_action AR_ACT_AxoclawShotMove2;
CK_action AR_ACT_AxoclawShotMove3;
CK_action AR_ACT_AxoclawShotMove4;

CK_action AR_ACT_AxoclawShotDie1;
CK_action AR_ACT_AxoclawShotDie2;

CK_action AR_ACT_AxoclawMove1 = {SPR_AXOCLAWMOVE1, SPR_AXOCLAWMOVE1, AT_ScaledOnce, 0, 0, 18, 10, 0, AR_AxoclawThink, CK_DeadlyCol, AR_AxoclawReact, &AR_ACT_AxoclawMove2};
CK_action AR_ACT_AxoclawMove2 = {SPR_AXOCLAWMOVE2, SPR_AXOCLAWMOVE2, AT_ScaledOnce, 0, 0, 18, 10, 0, AR_AxoclawThink, CK_DeadlyCol, AR_AxoclawReact, &AR_ACT_AxoclawMove1};


CK_action AR_ACT_AxoclawShoot1 = {SPR_AXOCLAWMOVE1, SPR_AXOCLAWMOVE1, AT_UnscaledFrame, 0, 0, 10, 0, 0, NULL, CK_DeadlyCol, AR_AxoclawReact, &AR_ACT_AxoclawShoot2};
CK_action AR_ACT_AxoclawShoot2 = {SPR_AXOCLAWMOVE2, SPR_AXOCLAWMOVE2, AT_UnscaledFrame, 0, 0, 10, 0, 0, NULL, CK_DeadlyCol, AR_AxoclawReact, &AR_ACT_AxoclawShoot3};
CK_action AR_ACT_AxoclawShoot3 = {SPR_AXOCLAWMOVE1, SPR_AXOCLAWMOVE1, AT_UnscaledFrame, 0, 0, 10, 0, 0, NULL, CK_DeadlyCol, AR_AxoclawReact, &AR_ACT_AxoclawShoot4};
CK_action AR_ACT_AxoclawShoot4 = {SPR_AXOCLAWMOVE2, SPR_AXOCLAWMOVE2, AT_UnscaledFrame, 0, 0, 10, 0, 0, NULL, CK_DeadlyCol, AR_AxoclawReact, &AR_ACT_AxoclawShoot5};
CK_action AR_ACT_AxoclawShoot5 = {SPR_AXOCLAWSHOOT1, SPR_AXOCLAWSHOOT1, AT_UnscaledOnce, 0, 0, 6, 0, 0, AR_AxoclawShoot, CK_DeadlyCol, AR_AxoclawReact, &AR_ACT_AxoclawShoot6};
CK_action AR_ACT_AxoclawShoot6 = {SPR_AXOCLAWSHOOT1, SPR_AXOCLAWSHOOT1, AT_UnscaledOnce, 0, 0, 42, 0, 0, NULL, CK_DeadlyCol, AR_AxoclawReact, &AR_ACT_AxoclawMove1};

CK_action AR_ACT_AxoclawShotMove1 = {SPR_AXOSHOT1, SPR_AXOSHOT1, AT_ScaledFrame, 0, 0, 6, 12, 0, AR_AxoclawShot, CK_DeadlyCol2, AR_AxoclawShotReact, &AR_ACT_AxoclawShotMove2};
CK_action AR_ACT_AxoclawShotMove2 = {SPR_AXOSHOT2, SPR_AXOSHOT2, AT_ScaledFrame, 0, 0, 6, 12, 0, AR_AxoclawShot, CK_DeadlyCol2, AR_AxoclawShotReact, &AR_ACT_AxoclawShotMove3};
CK_action AR_ACT_AxoclawShotMove3 = {SPR_AXOSHOT3, SPR_AXOSHOT3, AT_ScaledFrame, 0, 0, 6, 12, 0, AR_AxoclawShot, CK_DeadlyCol2, AR_AxoclawShotReact, &AR_ACT_AxoclawShotMove4};
CK_action AR_ACT_AxoclawShotMove4 = {SPR_AXOSHOT4, SPR_AXOSHOT4, AT_ScaledFrame, 0, 0, 6, 12, 0, AR_AxoclawShot, CK_DeadlyCol2, AR_AxoclawShotReact, &AR_ACT_AxoclawShotMove1};


CK_action AR_ACT_AxoclawShotDie1 = {SPR_AXOSHOTDIE1, SPR_AXOSHOTDIE1, AT_UnscaledFrame, 0, 0, 6, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_AxoclawShotDie2};
CK_action AR_ACT_AxoclawShotDie2 = {SPR_AXOSHOTDIE2, SPR_AXOSHOTDIE2, AT_UnscaledFrame, 0, 0, 6, 0, 0, NULL, NULL, CK_BasicDrawFunc1, NULL};

void AR_SpawnAxoclaw(int tileX, int tileY)
{
	CK_object *new_object = CK_GetNewObj(false);
	new_object->type = CT_Axoclaw;
	new_object->active = OBJ_ACTIVE;
	new_object->zLayer = 0;
	new_object->posX = tileX << 8;
	new_object->posY = (tileY << 8);  
	new_object->xDirection = US_RndT() < 0x80 ? motion_Left : motion_Right;
	new_object->yDirection = motion_None;
	new_object->clipped = CLIP_normal;
	CK_SetAction(new_object, &AR_ACT_AxoclawMove1);
}

void AR_AxoclawThink(CK_object *obj)
{
	int dx = ck_keenObj->clipRects.unitXmid - obj->clipRects.unitXmid;

	// Follow keen
	if (US_RndT() < 0x40)
	{
		if (dx > 0x400)
			obj->xDirection = motion_Right;
		else if (dx < -0x400)
			obj->xDirection = motion_Left;
	}

	// Switch to shooting state if keen in range
	if (ABS(dx) < 0x900 && ck_keenObj->posY > obj->posY && US_RndT() < 0x20)
	{
		CK_SetAction2(obj, &AR_ACT_AxoclawShoot1);
		return;
	}



}

void AR_SpawnAxoclawMissile(unsigned x, unsigned y, int dir)
{

	// Spawn a missile
	CK_object *new_object = CK_GetNewObj(true);
	new_object->type = CT_Axoclawshot;
	new_object->active = OBJ_EXISTS_ONLY_ONSCREEN;
	new_object->zLayer = 2;
	new_object->posX = x; 
	new_object->posY = y; 
	new_object->xDirection = dir;
	new_object->velY = 0;
	new_object->clipped = CLIP_normal;
	SD_PlaySound(AXOCLAWSHOOTSND);
	CK_SetAction(new_object, &AR_ACT_AxoclawShotMove1);


}

void AR_AxoclawShoot(CK_object *obj)
{
	int dir = ck_keenObj->clipRects.unitXmid > obj->clipRects.unitXmid ? motion_Right : motion_Left;
	AR_SpawnAxoclawMissile(obj->clipRects.unitXmid - 0x40, obj->posY + 0x120, dir);
}


void AR_AxoclawReact(CK_object *obj)
{

	if (obj->leftTI)
		obj->xDirection = motion_Left;

	if (obj->rightTI)
		obj->xDirection = motion_Right;

	RF_PlaceSprite(&obj->sde, obj->posX, obj->posY, obj->gfxChunk, false, obj->zLayer);
	
}


void AR_AxoclawShot(CK_object *obj)
{

	int dy = ck_keenObj->posY - obj->posY;

	if (dy > 0)
		obj->yDirection = motion_Down;
	else
		obj->yDirection = motion_Up;


	CK_PhysAccelVert(obj, obj->yDirection, 4);

}


void AR_AxoclawShotReact(CK_object *obj)
{

	if (obj->bottomTI || obj->topTI || obj->leftTI || obj->rightTI)
	{
		CK_SetAction2(obj, &AR_ACT_AxoclawShotDie1);
	}

	
	RF_PlaceSprite(&obj->sde, obj->posX, obj->posY, obj->gfxChunk, false, obj->zLayer);
}

//===========================================================================
//
// Axosis and Krile Ghost (Smirky / Arachnut Replacement)
// 
//===========================================================================

// Description
// The Axosis walks back and forth and is deadly to touch.  It only vulnerable if contacted by a Krile Ghost.
// If the Krile Ghost touches the Axosis, then the Axosis is stunned and vulnerable to Keen's pogo stick.

// The Krile ghost teleports around the level, alternating between random locations and locations near the top left
// axosis.

// Ghost searches for an appropriately large space to warp in, then teleports and falls to floor
// If in attack mode, warps to next axosis; otherwise, just roams around for awhile
// Keen must be in visual range of the top left axosis for the ghost to warp and attack it

CK_action AR_ACT_AxosisWalk1;
CK_action AR_ACT_AxosisWalk2;

CK_action AR_ACT_AxosisDead1;

CK_action AR_ACT_AxosisStunned1;
CK_action AR_ACT_AxosisStunned2;
CK_action AR_ACT_AxosisStunned3;
CK_action AR_ACT_AxosisStunned4;
CK_action AR_ACT_AxosisStunned5;
CK_action AR_ACT_AxosisStunned6;
CK_action AR_ACT_AxosisStunned7;
CK_action AR_ACT_AxosisStunned8;
CK_action AR_ACT_AxosisStunned9;
CK_action AR_ACT_AxosisStunned10;
CK_action AR_ACT_AxosisStunned11;
CK_action AR_ACT_AxosisStunned12;
CK_action AR_ACT_AxosisStunned13;

void AR_SpawnAxosis(int tileX, int tileY);
void AR_AxosisWalk(CK_object *obj);
void AR_AxosisStunCount(CK_object *obj);
void AR_AxosisCollide(CK_object *obj1, CK_object *obj2);
void AR_AxosisStunnedCollide(CK_object *obj1, CK_object *obj2);

#define KRILEGHOSTSTUNDELAY 400
CK_action AR_ACT_KrileGhostHide;

CK_action AR_ACT_KrileGhostRoam1;
CK_action AR_ACT_KrileGhostRoam2;

CK_action AR_ACT_KrileGhostTeleIn1;
CK_action AR_ACT_KrileGhostTeleIn2;
CK_action AR_ACT_KrileGhostTeleIn3;
CK_action AR_ACT_KrileGhostTeleIn4;
CK_action AR_ACT_KrileGhostTeleOut1;
CK_action AR_ACT_KrileGhostTeleOut2;
CK_action AR_ACT_KrileGhostTeleOut3;
CK_action AR_ACT_KrileGhostTeleOut4;

CK_action AR_ACT_KrileGhostHunt1;
CK_action AR_ACT_KrileGhostHunt2;

CK_action AR_ACT_KrileGhostZap1;
CK_action AR_ACT_KrileGhostZap2;
CK_action AR_ACT_KrileGhostZap3;
CK_action AR_ACT_KrileGhostZap4;
CK_action AR_ACT_KrileGhostZap5;
CK_action AR_ACT_KrileGhostZap6;
CK_action AR_ACT_KrileGhostZap7;
CK_action AR_ACT_KrileGhostZap8;
CK_action AR_ACT_KrileGhostTeleOut5;

void AR_SpawnKrileGhost(int tileX, int tileY);
void AR_KrileGhostHide(CK_object *obj);
void AR_KrileGhostFindAxosis(CK_object *obj);
void AR_KrileGhostHunt(CK_object *obj);
void AR_KrileGhostHuntCollide(CK_object *obj1, CK_object *obj2);
void AR_KrileGhostRoamReact(CK_object *obj);





CK_action AR_ACT_AxosisWalk1 = {AXOSISWALKL1SPR, AXOSISWALKR1SPR, AT_ScaledOnce, 0, 1, 8, 20, 0, AR_AxosisWalk, AR_AxosisCollide, CK_BasicDrawFunc2, &AR_ACT_AxosisWalk2};
CK_action AR_ACT_AxosisWalk2 = {AXOSISWALKL2SPR, AXOSISWALKR2SPR, AT_ScaledOnce, 0, 1, 8, 20, 0, AR_AxosisWalk, AR_AxosisCollide, CK_BasicDrawFunc2, &AR_ACT_AxosisWalk1};

CK_action AR_ACT_AxosisDead1 = {SPR_AXOSISDEAD, SPR_AXOSISDEAD, AT_Frame, 0, 0, 8, 0, 0, CK_Fall, NULL, CK_BasicDrawFunc1, &AR_ACT_AxosisDead1};

CK_action AR_ACT_AxosisStunned1 = {SPR_AXOSISSTUN1, SPR_AXOSISSTUN1, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, AR_AxosisStunnedCollide, CK_BasicDrawFunc1, &AR_ACT_AxosisStunned2};
CK_action AR_ACT_AxosisStunned2 = {SPR_AXOSISSTUN2, SPR_AXOSISSTUN2, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, AR_AxosisStunnedCollide, CK_BasicDrawFunc1, &AR_ACT_AxosisStunned3};
CK_action AR_ACT_AxosisStunned3 = {SPR_AXOSISSTUN3, SPR_AXOSISSTUN3, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, AR_AxosisStunnedCollide, CK_BasicDrawFunc1, &AR_ACT_AxosisStunned4};
CK_action AR_ACT_AxosisStunned4 = {SPR_AXOSISSTUN4, SPR_AXOSISSTUN4, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, AR_AxosisStunnedCollide, CK_BasicDrawFunc1, &AR_ACT_AxosisStunned5};
CK_action AR_ACT_AxosisStunned5 = {SPR_AXOSISSTUNNED, SPR_AXOSISSTUNNED, AT_UnscaledOnce, 0, 0, 180, 0, 0, AR_AxosisStunCount, AR_AxosisStunnedCollide, CK_BasicDrawFunc1, &AR_ACT_AxosisStunned6};
CK_action AR_ACT_AxosisStunned6 = {SPR_AXOSISPREEN1, SPR_AXOSISPREEN1, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, AR_AxosisStunnedCollide, CK_BasicDrawFunc1, &AR_ACT_AxosisStunned7};
CK_action AR_ACT_AxosisStunned7 = {SPR_AXOSISPREEN2, SPR_AXOSISPREEN2, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, AR_AxosisStunnedCollide, CK_BasicDrawFunc1, &AR_ACT_AxosisStunned8};
CK_action AR_ACT_AxosisStunned8 = {SPR_AXOSISPREEN3, SPR_AXOSISPREEN3, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, AR_AxosisStunnedCollide, CK_BasicDrawFunc1, &AR_ACT_AxosisStunned9};
CK_action AR_ACT_AxosisStunned9 = {SPR_AXOSISPREEN4, SPR_AXOSISPREEN4, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, AR_AxosisStunnedCollide, CK_BasicDrawFunc1, &AR_ACT_AxosisStunned5};
CK_action AR_ACT_AxosisStunned10 = {SPR_AXOSISSTUN4, SPR_AXOSISSTUN4, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, AR_AxosisStunnedCollide, CK_BasicDrawFunc1, &AR_ACT_AxosisStunned11};
CK_action AR_ACT_AxosisStunned11 = {SPR_AXOSISSTUN3, SPR_AXOSISSTUN3, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, AR_AxosisStunnedCollide, CK_BasicDrawFunc1, &AR_ACT_AxosisStunned12};
CK_action AR_ACT_AxosisStunned12 = {SPR_AXOSISSTUN2, SPR_AXOSISSTUN2, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, AR_AxosisStunnedCollide, CK_BasicDrawFunc1, &AR_ACT_AxosisStunned13};
CK_action AR_ACT_AxosisStunned13 = {SPR_AXOSISSTUN1, SPR_AXOSISSTUN1, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, AR_AxosisStunnedCollide, CK_BasicDrawFunc1, &AR_ACT_AxosisWalk1};

// user1; holds index in list of Axosises

void AR_SpawnAxosis(int tileX, int tileY)
{
	CK_object *new_object = CK_GetNewObj(false);
	new_object->type = CT_Axosis;
	new_object->active = OBJ_ACTIVE;
	new_object->zLayer = 0;
	new_object->posX = tileX << 8;
	new_object->posY = (tileY << 8) - 0x180;  
	new_object->xDirection = US_RndT() < 0x80 ? motion_Left : motion_Right;
	new_object->yDirection = motion_None;
	new_object->clipped = CLIP_normal;
	CK_SetAction(new_object, &AR_ACT_AxosisWalk1);

	new_object->user1 = ck_gameState.axosisNum++;
}


void AR_AxosisWalk(CK_object *obj)
{
	int dx = ck_keenObj->posX - obj->posX;

	// Seek keen
	if (US_RndT() < 0x8)
		obj->xDirection = dx > 0 ? motion_Right : motion_Left;

}

void AR_AxosisStunCount(CK_object *obj)
{
	if (obj->user2++ == 3)
	{
		obj->user2 = 0;
		CK_SetAction2(obj, &AR_ACT_AxosisStunned10);
	}

}

void AR_AxosisCollide(CK_object *obj1, CK_object *obj2)
{
	if (obj2->type == CT_Player)
	{
		CK_KillKeen();
		return;
	}

	if (obj2->type == CT_Stunner)
	{
		CK_ShotHit(obj2);
		return;
	}

}

void AR_AxosisStunnedCollide(CK_object *obj1, CK_object *obj2)
{
	// K1n9_Duk3 fix: nextAxosis is not used in here
	/*
	CK_object *nextAxosis;

	if (ck_gameState.krileGhost)
		nextAxosis = (CK_object *)ck_gameState.krileGhost->user4;
	*/
		
	// Squish Axosis if keen is pogoing
	if (obj2->type == CT_Player)
	{
		if (obj2->currentAction >= &CK_ACT_keenPogo1 && obj2->currentAction <= &CK_ACT_keenPogo3)
		{
			// Play a squish sound
			SD_PlaySound(AXOSISSQUISHSND);
			
			CK_SetAction2(obj1, &AR_ACT_AxosisDead1);

			// Tell the ghost to find the next axosis
			if (ck_gameState.krileGhost)
			{
				// K1n9_Duk3 fix:
				// This makes sure the ghost is teleporting out, so it won't
				// get stuck near the dead axosis. You could just change the
				// ghost's action to AR_ACT_KrileGhostTeleOut5, but matching
				// the teleport state to the current sprite looks much better.
				{
					register CK_action *nextaction;

					switch (ck_gameState.krileGhost->gfxChunk)
					{
					case 0:
					case -1:
						nextaction = &AR_ACT_KrileGhostTeleOut5;
						break;
					case SPR_KRILEGHOSTTELE4:
						nextaction = &AR_ACT_KrileGhostTeleOut4;
						break;
					case SPR_KRILEGHOSTTELE3:
						nextaction = &AR_ACT_KrileGhostTeleOut3;
						break;
					case SPR_KRILEGHOSTTELE2:
						nextaction = &AR_ACT_KrileGhostTeleOut2;
						break;
					default:
						nextaction = &AR_ACT_KrileGhostTeleOut1;
						break;
					}

					CK_SetAction2(ck_gameState.krileGhost, nextaction);
				}

				if (++(ck_gameState.krileGhost->user3) < ck_gameState.axosisNum)
				{
					CK_object *o;
					for (o = ck_keenObj; o; o = o->next)
					{
						if (o->type == CT_Axosis && o->user1 == ck_gameState.krileGhost->user3)
						{
							// HACK: Ghosts are reviving the last stunned axosis
							// Therefore, explicitly tell the ghost to ignore dead axosises
							// and hope that stops it from happening

							// K1n9_Duk3 mod: don't need the check anymore
							/*
							if (o->currentAction == &AR_ACT_AxosisDead1)
							{
								ck_gameState.krileGhost->user4 = NULL;
								continue;
							}
							*/

							ck_gameState.krileGhost->user4 = (unsigned)o;
							return;
						}
					}
				}
				// K1n9_Duk3 fix: removed the else and added a return in the loop
				// to make sure user4 is set to NULL when no Axosis is found.

				// No more axosises
				ck_gameState.krileGhost->user4 = NULL;

			}

			return;
		}
	}
}


CK_action AR_ACT_KrileGhostHide = {-1, -1, AT_Frame, 0, 0, 8, 0, 0, AR_KrileGhostHide, NULL, CK_BasicDrawFunc1, &AR_ACT_KrileGhostHide};

CK_action AR_ACT_KrileGhostMoveToAxosis = {SPR_KRILEGHOSTFLOAT1, SPR_KRILEGHOSTFLOAT1, AT_UnscaledOnce, 0, 0, 140, 0, 0, AR_KrileGhostFindAxosis, NULL, NULL, &AR_ACT_KrileGhostTeleIn1};

CK_action AR_ACT_KrileGhostTeleIn1 = {SPR_KRILEGHOSTTELE4, SPR_KRILEGHOSTTELE4, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_KrileGhostTeleIn2};
CK_action AR_ACT_KrileGhostTeleIn2 = {SPR_KRILEGHOSTTELE3, SPR_KRILEGHOSTTELE3, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_KrileGhostTeleIn3};
CK_action AR_ACT_KrileGhostTeleIn3 = {SPR_KRILEGHOSTTELE2, SPR_KRILEGHOSTTELE2, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_KrileGhostTeleIn4};
CK_action AR_ACT_KrileGhostTeleIn4 = {SPR_KRILEGHOSTTELE1, SPR_KRILEGHOSTTELE1, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_KrileGhostHunt1};

CK_action AR_ACT_KrileGhostTeleOut1 = {SPR_KRILEGHOSTTELE1, SPR_KRILEGHOSTTELE1, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_KrileGhostTeleOut2};
CK_action AR_ACT_KrileGhostTeleOut2 = {SPR_KRILEGHOSTTELE2, SPR_KRILEGHOSTTELE2, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_KrileGhostTeleOut3};
CK_action AR_ACT_KrileGhostTeleOut3 = {SPR_KRILEGHOSTTELE3, SPR_KRILEGHOSTTELE3, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_KrileGhostTeleOut4};
CK_action AR_ACT_KrileGhostTeleOut4 = {SPR_KRILEGHOSTTELE4, SPR_KRILEGHOSTTELE4, AT_UnscaledOnce, 0, 0, 8, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_KrileGhostTeleOut5};
CK_action AR_ACT_KrileGhostTeleOut5 = {-1, -1, AT_UnscaledOnce, 0, 1, 600, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_KrileGhostHide};

CK_action AR_ACT_KrileGhostHunt1 = {SPR_KRILEGHOSTFLOAT1, SPR_KRILEGHOSTFLOAT1, AT_ScaledFrame, 0, 1, 6, 4, 0, AR_KrileGhostHunt, AR_KrileGhostHuntCollide, AR_KrileGhostRoamReact, &AR_ACT_KrileGhostHunt2};
CK_action AR_ACT_KrileGhostHunt2 = {SPR_KRILEGHOSTFLOAT2, SPR_KRILEGHOSTFLOAT2, AT_ScaledFrame, 0, 1, 6, 4, 0, AR_KrileGhostHunt, AR_KrileGhostHuntCollide, AR_KrileGhostRoamReact, &AR_ACT_KrileGhostHunt1};

CK_action AR_ACT_KrileGhostZap1 = {SPR_KRILEGHOSTATTACK1, SPR_KRILEGHOSTATTACK1, AT_UnscaledOnce, 0, 1, 6, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_KrileGhostZap2};
CK_action AR_ACT_KrileGhostZap2 = {SPR_KRILEGHOSTATTACK2, SPR_KRILEGHOSTATTACK2, AT_UnscaledOnce, 0, 1, 6, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_KrileGhostZap3};
CK_action AR_ACT_KrileGhostZap3 = {SPR_KRILEGHOSTATTACK1, SPR_KRILEGHOSTATTACK1, AT_UnscaledOnce, 0, 1, 6, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_KrileGhostZap4};
CK_action AR_ACT_KrileGhostZap4 = {SPR_KRILEGHOSTATTACK2, SPR_KRILEGHOSTATTACK2, AT_UnscaledOnce, 0, 1, 6, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_KrileGhostZap5};
CK_action AR_ACT_KrileGhostZap5 = {SPR_KRILEGHOSTATTACK1, SPR_KRILEGHOSTATTACK1, AT_UnscaledOnce, 0, 1, 6, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_KrileGhostZap6};
CK_action AR_ACT_KrileGhostZap6 = {SPR_KRILEGHOSTATTACK2, SPR_KRILEGHOSTATTACK2, AT_UnscaledOnce, 0, 1, 6, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_KrileGhostZap7};
CK_action AR_ACT_KrileGhostZap7 = {SPR_KRILEGHOSTATTACK1, SPR_KRILEGHOSTATTACK1, AT_UnscaledOnce, 0, 1, 6, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_KrileGhostZap8};
CK_action AR_ACT_KrileGhostZap8 = {SPR_KRILEGHOSTATTACK2, SPR_KRILEGHOSTATTACK2, AT_UnscaledOnce, 0, 1, 6, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_KrileGhostTeleOut1};

//user3; index of Axosis to hunt;
//user4: pointer to said axosis; if NULL, then don't search 

void AR_SpawnKrileGhost(int tileX, int tileY)
{
	CK_object *new_object;

	if (ck_gameState.krileGhost)
		Quit("More than one Krile ghost!");

	new_object = CK_GetNewObj(false);
	new_object->type = CT_KrileGhost;
	new_object->active = OBJ_ALWAYS_ACTIVE;
	new_object->zLayer = 1;
	new_object->posX = tileX << 8;
	new_object->posY = (tileY << 8) - 0x300;  
	new_object->xDirection = US_RndT() < 0x80 ? motion_Left : motion_Right;
	new_object->yDirection = motion_None;
	new_object->clipped = CLIP_full;
	CK_SetAction(new_object, &AR_ACT_KrileGhostHide);
	ck_gameState.krileGhost = new_object;
	
}

void AR_KrileGhostHide(CK_object *obj)
{
	CK_object *nextAxosis = (CK_object *)obj->user4;

	if (nextAxosis)
	{
		int dx = nextAxosis->posX - ck_keenObj->posX;
		int dy = nextAxosis->posY - ck_keenObj->posY;
		if (dy > -0x200 && dy < 0x900 && dx > -0x600 && dx < 0x600)
			CK_SetAction2(obj, &AR_ACT_KrileGhostMoveToAxosis);
	}
	else
	{
		CK_RemoveObj(obj);
	}
}

void AR_KrileGhostFindAxosis(CK_object *obj)
{
	CK_object *nextAxosis = (CK_object *)obj->user4;

	if (!nextAxosis)
	{
		CK_RemoveObj(obj);
		return;
	}
	// Move to Ghost coords

	obj->posX = nextAxosis->posX;
	obj->posY = nextAxosis->posY - 0x180;
	obj->user2 = KRILEGHOSTSTUNDELAY;
	CK_SetAction(obj, &AR_ACT_KrileGhostTeleIn1);
}

void AR_KrileGhostHunt(CK_object *obj)
{
	CK_object *nextAxosis = (CK_object *)obj->user4;

	if (!nextAxosis)
	{
		CK_RemoveObj(obj);
		return;
	}

	// Track down the next Axosis
	if (((int)obj->user2 -= tics)< 0)
		obj->user2 = 0;

	if (nextAxosis && obj->user2 < KRILEGHOSTSTUNDELAY/2)
	{
		obj->xDirection = nextAxosis->posX > obj->posX? motion_Right : motion_Left;
	}

}

void AR_KrileGhostHuntCollide(CK_object *obj1, CK_object *obj2)
{

	CK_object *nextAxosis = (CK_object *)obj1->user4;

	if (obj2 && obj2 == nextAxosis && (int)obj1->user2 == 0)
	{
		// make sure it's centered
		int dx = obj1->clipRects.unitXmid - obj2->clipRects.unitXmid;
		if (ABS(dx) < 0x40)
		{
			CK_SetAction2(obj2, &AR_ACT_AxosisStunned1);
			CK_SetAction2(obj1, &AR_ACT_KrileGhostZap1);
			obj1->user2 = KRILEGHOSTSTUNDELAY;
			SD_PlaySound(KRILEGHOSTATTACKSND);
			return;
		}
	}
}

void AR_KrileGhostRoamReact(CK_object *obj)
{

	// Hit wall walking right; turn around and go left
	if (obj->xDirection == motion_Right && obj->leftTI != 0)
	{
		obj->posX -= obj->deltaPosX;
		obj->xDirection = motion_Left;
		obj->timeUntillThink = US_RndT() / 32;
		CK_SetAction2(obj, obj->currentAction);
	}
		// Hit wall walking left; turn around and go right
	else if (obj->xDirection == motion_Left && obj->rightTI != 0)
	{
		obj->posX -= obj->deltaPosX;
		obj->xDirection = motion_Right;
		obj->timeUntillThink = US_RndT() / 32;
		CK_SetAction2(obj, obj->currentAction);
	}
		// Walked off of ledge; turn around
	else if (obj->topTI == 0)
	{
		obj->posX -= obj->deltaPosX;
		obj->xDirection = -obj->xDirection;
		obj->timeUntillThink = US_RndT() / 32;
		CK_SetAction2(obj, obj->currentAction);
	}

	RF_PlaceSprite(&obj->sde, obj->posX, obj->posY, obj->gfxChunk, US_RndT()>0xC0?maskdraw:spritedraw, obj->zLayer);
	
}



//===========================================================================
//
// MOXEL (Sprite replacement)
// 
//===========================================================================

// Description: An immobile device that sits on vertical walls and launches fireballs if keen passes before it
// It always points to the left.

CK_action AR_ACT_MoxelWait;
CK_action AR_ACT_MoxelWaitToShoot;
CK_action AR_ACT_MoxelShoot;
CK_action AR_ACT_MoxelShoot2;

CK_action AR_ACT_MoxelShotFall1;
CK_action AR_ACT_MoxelShotFall2;

CK_action AR_ACT_MoxelShotGround1;
CK_action AR_ACT_MoxelShotGround2;

void AR_SpawnMoxel(int tileX, int tileY);
boolean AR_KeenWithinMoxelRange(CK_object *obj);
void AR_MoxelThink(CK_object *obj);
void AR_MoxelWaitToShoot(CK_object *obj);
void AR_MoxelShotBurnout(CK_object *obj);
void AR_MoxelShotReact(CK_object *obj);

CK_action AR_ACT_MoxelWait = {SPR_MOXEL1, SPR_MOXEL1, AT_Frame, 0, 0, 8, 0, 0, AR_MoxelThink, NULL, CK_BasicDrawFunc1, NULL};
CK_action AR_ACT_MoxelWaitToShoot = {SPR_MOXELSHOOT1, SPR_MOXELSHOOT1, AT_Frame, 0, 0, 55, 0, 0, AR_MoxelWaitToShoot, NULL, CK_BasicDrawFunc1, NULL};
CK_action AR_ACT_MoxelShoot = {SPR_MOXELSHOOT2, SPR_MOXELSHOOT2, AT_UnscaledOnce, 0, 0, 35, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_MoxelShoot2};
CK_action AR_ACT_MoxelShoot2 = {SPR_MOXEL1, SPR_MOXEL1, AT_UnscaledOnce, 0, 0, 90, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_MoxelWait};

CK_action AR_ACT_MoxelShotFall1 = {SPR_MOXELSHOT1, SPR_MOXELSHOT1, AT_ScaledFrame, 0, 0, 6, 0, 0, CK_Fall, CK_DeadlyCol2, AR_MoxelShotReact, &AR_ACT_MoxelShotFall2};
CK_action AR_ACT_MoxelShotFall2 = {SPR_MOXELSHOT2, SPR_MOXELSHOT2, AT_ScaledFrame, 0, 0, 6, 0, 0, CK_Fall, CK_DeadlyCol2, AR_MoxelShotReact, &AR_ACT_MoxelShotFall1};

CK_action AR_ACT_MoxelShotGround1 = {SPR_MOXELSHOT3, SPR_MOXELSHOT3, AT_ScaledFrame, 0, 0, 6, 0, 0, AR_MoxelShotBurnout, CK_DeadlyCol2, CK_BasicDrawFunc1, &AR_ACT_MoxelShotGround2};  
CK_action AR_ACT_MoxelShotGround2 = {SPR_MOXELSHOT4, SPR_MOXELSHOT4, AT_ScaledFrame, 0, 0, 6, 0, 0, AR_MoxelShotBurnout, CK_DeadlyCol2, CK_BasicDrawFunc1, &AR_ACT_MoxelShotGround1};  

void AR_SpawnMoxel(int tileX, int tileY)
{
	CK_object *new_object = CK_GetNewObj(false);
	new_object->type = CT_Moxel;
	new_object->active = OBJ_ACTIVE;
	new_object->zLayer = 0;
	new_object->posX = tileX << 8;
	new_object->posY = (tileY << 8);  
	new_object->xDirection = 0;
	new_object->yDirection = motion_None;
	new_object->clipped = CLIP_not;
	CK_SetAction(new_object, &AR_ACT_MoxelWait);
}

boolean AR_KeenWithinMoxelRange(CK_object *obj)
{
	int dx = ck_keenObj->posX - obj->posX;
	int dy = ck_keenObj->posY - obj->posY;

	if (dx > -0x980 && dx < 0 && dy < 0x200 && dy > -0x200)
	{
		return true;
	}

	return false;

}

void AR_MoxelThink(CK_object *obj)
{

	// Switch to throwing state is keen is in front of the moxel
	if (AR_KeenWithinMoxelRange(obj))
	{
		CK_SetAction2(obj, &AR_ACT_MoxelWaitToShoot);
		obj->user1 = 55;
		return;
	}

}

void AR_MoxelWaitToShoot(CK_object *obj)
{
	if (AR_KeenWithinMoxelRange(obj))
	{
		if ((obj->user1 -= tics) < 0) 
		{
			// Spawn a missile
			CK_object *new_object = CK_GetNewObj(true);
			new_object->type = CT_MoxelShot;
			new_object->active = OBJ_EXISTS_ONLY_ONSCREEN;
			new_object->zLayer = 2;
			new_object->posX = obj->posX - 0x80;
			new_object->posY = obj->posY + 0x40;
			new_object->xDirection = motion_Left;
			new_object->velX = -40;
			new_object->velY = -20;
			new_object->clipped = CLIP_normal;
			CK_SetAction(new_object, &AR_ACT_MoxelShotFall1);

			// 
			SD_PlaySound(SENTRYSHOOTSND);
			CK_SetAction2(obj, &AR_ACT_MoxelShoot);
			return;
		}
	}
	else
	{
		// Reset
		CK_SetAction2(obj, &AR_ACT_MoxelWait);
		return;
	}
}
	

void AR_MoxelShotBurnout(CK_object *obj)
{
	if ((obj->user1 -= tics) < 0)
	{
		CK_RemoveObj(obj);
		return;
	}

}

void AR_MoxelShotReact(CK_object *obj)
{
	if (obj->topTI)
	{
		CK_SetAction2(obj, &AR_ACT_MoxelShotGround1);
		obj->user1 = 70;
		SD_PlaySound(SENTRYSHOTIMPACTSND);
		return;
	}

	if (obj->leftTI || obj->rightTI)
	{
		obj->velX = 0;
	}

	RF_PlaceSprite(&obj->sde, obj->posX, obj->posY, obj->gfxChunk, false, obj->zLayer);

}


//===========================================================================
//
// Dark Krile Guardian (Berkeloid Replacement)
// 
//===========================================================================

// Description: A berkeloid that shoots projectiles in a straight line.
// Deadly to keen, turns around if it detects keen

CK_action AR_ACT_KrileGuardianWalk1;
CK_action AR_ACT_KrileGuardianWalk2;
CK_action AR_ACT_KrileGuardianWalk3;
CK_action AR_ACT_KrileGuardianWalk4;
CK_action AR_ACT_KrileGuardianWalk5;
CK_action AR_ACT_KrileGuardianWalk6;

CK_action AR_ACT_KrileGuardianShoot1;
CK_action AR_ACT_KrileGuardianShoot2;

CK_action AR_ACT_KrileGuardShotMove1;
CK_action AR_ACT_KrileGuardShotMove2;
CK_action AR_ACT_KrileGuardShotMove3;
CK_action AR_ACT_KrileGuardShotMove4;

CK_action AR_ACT_KrileGuardShotDie1;
CK_action AR_ACT_KrileGuardShotDie2;
CK_action AR_ACT_KrileGuardShotDie3;


void AR_SpawnKrileGuardian(int tileX, int tileY);
void AR_KrileGuardianWalk(CK_object *obj);
void AR_KrileGuardianShoot(CK_object *obj);
void AR_KrileGuardianReact(CK_object *obj);
void AR_KrileGuardianShotReact(CK_object *obj);

CK_action AR_ACT_KrileGuardianWalk1 = {SPR_KRILEGUARDWALKL1, SPR_KRILEGUARDWALKR1, AT_ScaledOnce, 0, 1, 8, 16, 0, AR_KrileGuardianWalk, CK_DeadlyCol, AR_KrileGuardianReact, &AR_ACT_KrileGuardianWalk2};
CK_action AR_ACT_KrileGuardianWalk2 = {SPR_KRILEGUARDWALKL2, SPR_KRILEGUARDWALKR2, AT_ScaledOnce, 0, 1, 8, 16, 0, AR_KrileGuardianWalk, CK_DeadlyCol, AR_KrileGuardianReact, &AR_ACT_KrileGuardianWalk3};
CK_action AR_ACT_KrileGuardianWalk3 = {SPR_KRILEGUARDWALKL3, SPR_KRILEGUARDWALKR3, AT_ScaledOnce, 0, 1, 8, 16, 0, AR_KrileGuardianWalk, CK_DeadlyCol, AR_KrileGuardianReact, &AR_ACT_KrileGuardianWalk4};
CK_action AR_ACT_KrileGuardianWalk4 = {SPR_KRILEGUARDWALKL4, SPR_KRILEGUARDWALKR4, AT_ScaledOnce, 0, 1, 8, 16, 0, AR_KrileGuardianWalk, CK_DeadlyCol, AR_KrileGuardianReact, &AR_ACT_KrileGuardianWalk5};
CK_action AR_ACT_KrileGuardianWalk5 = {SPR_KRILEGUARDWALKL3, SPR_KRILEGUARDWALKR3, AT_ScaledOnce, 0, 1, 8, 16, 0, AR_KrileGuardianWalk, CK_DeadlyCol, AR_KrileGuardianReact, &AR_ACT_KrileGuardianWalk6};
CK_action AR_ACT_KrileGuardianWalk6 = {SPR_KRILEGUARDWALKL2, SPR_KRILEGUARDWALKR2, AT_ScaledOnce, 0, 1, 8, 16, 0, AR_KrileGuardianWalk, CK_DeadlyCol, AR_KrileGuardianReact, &AR_ACT_KrileGuardianWalk1};

CK_action AR_ACT_KrileGuardianShoot1 = {SPR_KRILEGUARDSHOOTL1, SPR_KRILEGUARDSHOOTR1, AT_ScaledOnce, 0, 1, 24, 16, 0, AR_KrileGuardianShoot, CK_DeadlyCol, AR_KrileGuardianReact, &AR_ACT_KrileGuardianShoot2};
CK_action AR_ACT_KrileGuardianShoot2 = {SPR_KRILEGUARDSHOOTL2, SPR_KRILEGUARDSHOOTR2, AT_ScaledOnce, 0, 1, 24, 16, 0, NULL, CK_DeadlyCol, AR_KrileGuardianReact, &AR_ACT_KrileGuardianWalk1};

CK_action AR_ACT_KrileGuardShotMove1 = {SPR_GUARDSHOT1, SPR_GUARDSHOT1, AT_ScaledOnce, 0, 0, 3, 40, 0, NULL, CK_DeadlyCol2, AR_KrileGuardianShotReact, &AR_ACT_KrileGuardShotMove2};
CK_action AR_ACT_KrileGuardShotMove2 = {SPR_GUARDSHOT2, SPR_GUARDSHOT2, AT_ScaledOnce, 0, 0, 3, 40, 0, NULL, CK_DeadlyCol2, AR_KrileGuardianShotReact, &AR_ACT_KrileGuardShotMove3};
CK_action AR_ACT_KrileGuardShotMove3 = {SPR_GUARDSHOT3, SPR_GUARDSHOT3, AT_ScaledOnce, 0, 0, 3, 40, 0, NULL, CK_DeadlyCol2, AR_KrileGuardianShotReact, &AR_ACT_KrileGuardShotMove4};
CK_action AR_ACT_KrileGuardShotMove4 = {SPR_GUARDSHOT4, SPR_GUARDSHOT4, AT_ScaledOnce, 0, 0, 3, 40, 0, NULL, CK_DeadlyCol2, AR_KrileGuardianShotReact, &AR_ACT_KrileGuardShotMove1};

CK_action AR_ACT_KrileGuardShotDie1 = {SPR_GUARDSHOTDIE1, SPR_GUARDSHOTDIE1, AT_ScaledOnce, 0, 0, 6, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_KrileGuardShotDie2};
CK_action AR_ACT_KrileGuardShotDie2 = {SPR_GUARDSHOTDIE2, SPR_GUARDSHOTDIE2, AT_ScaledOnce, 0, 0, 6, 0, 0, NULL, NULL, CK_BasicDrawFunc1, NULL};

void AR_SpawnKrileGuardian(int tileX, int tileY)
{
	CK_object *new_object = CK_GetNewObj(false);
	new_object->type = CT_KrileGuardian;
	new_object->active = OBJ_ACTIVE;
	new_object->zLayer = 0;
	new_object->posX = tileX << 8;
	new_object->posY = (tileY << 8) - 0x500;  
	new_object->xDirection = US_RndT() < 0x80 ? motion_Left : motion_Right;
	new_object->yDirection = motion_None;
	new_object->clipped = CLIP_normal;
	new_object->user2 = 8;
	CK_SetAction(new_object, &AR_ACT_KrileGuardianWalk1);

}

void AR_KrileGuardianWalk(CK_object *obj)
{

#if 0
	// Search for keen
	if (US_RndT() < 0x4)
	{
		int dx = ck_keenObj->clipRects.unitXmid - obj->clipRects.unitXmid;
		obj->xDirection = dx > 0? motion_Right : motion_Left;
	}
#endif

	// Shoot randomly
	if (US_RndT() < 12)
	{
		CK_SetAction2(obj, &AR_ACT_KrileGuardianShoot1);
		return;
	}
}

void AR_KrileGuardianShoot(CK_object *obj)
{
	// Spawn a missile
	CK_object *new_object = CK_GetNewObj(true);
	new_object->type = CT_GuardShot;
	new_object->active = OBJ_EXISTS_ONLY_ONSCREEN;
	new_object->zLayer = PRIORITIES - 2;
	new_object->posX = obj->clipRects.unitXmid + (obj->xDirection << 5);
	new_object->posY = obj->posY + 0x280;
	new_object->xDirection = obj->xDirection;
	new_object->yDirection = motion_None;
	new_object->clipped = CLIP_normal;
	CK_SetAction(new_object, &AR_ACT_KrileGuardShotMove1);
	SD_PlaySound(KRILEGUARDTHROWSND);
}

void AR_KrileGuardianReact(CK_object *obj)
{
	// Hit wall walking right; turn around and go left
	if (obj->xDirection == motion_Right && obj->leftTI != 0)
	{
		obj->posX -= obj->deltaPosX;
		obj->xDirection = motion_Left;
		obj->timeUntillThink = US_RndT() / 32;
		CK_SetAction2(obj, obj->currentAction);
	}
		// Hit wall walking left; turn around and go right
	else if (obj->xDirection == motion_Left && obj->rightTI != 0)
	{
		obj->posX -= obj->deltaPosX;
		obj->xDirection = motion_Right;
		obj->timeUntillThink = US_RndT() / 32;
		CK_SetAction2(obj, obj->currentAction);
	}
		// Walked off of ledge; turn around
	else if (obj->topTI == 0)
	{
		obj->posX -= obj->deltaPosX;
		obj->xDirection = -obj->xDirection;
		obj->timeUntillThink = US_RndT() / 32;
		CK_SetAction2(obj, obj->currentAction);
	}

	// Hovering action
	obj->user1 += (int)((int)obj->user2 * (int)tics);
	if ((int)obj->user1 > 0)
	{

		obj->user1 = 0;
		obj->user2 = -8;
	}
	else if ((int)obj->user1 < -0x100)
	{
		obj->user1 = -0x100;
		obj->user2 = 8;
	}
	
	RF_PlaceSprite(&obj->sde, obj->posX, obj->posY + obj->user1, obj->gfxChunk, false, obj->zLayer);
}

void AR_KrileGuardianShotReact(CK_object *obj)
{
	if (obj->topTI || obj->bottomTI || obj->leftTI || obj->rightTI)
	{
		// obj->posY += 0x140;
		obj->clipped = CLIP_not;
		CK_SetAction2(obj, &AR_ACT_KrileGuardShotDie1);
		return;
	}

	RF_PlaceSprite(&obj->sde, obj->posX, obj->posY, obj->gfxChunk, false, obj->zLayer);
}

//===========================================================================
//
// GENERATOR CORE 
// 
//===========================================================================

// description: Sits there, generating, untill Keen shoots it.  Then it blows up,
// displays a textbox, and the level ends

// IF KEEN DIES before it blows up, then no textbox and levelend are performed
CK_action AR_ACT_GeneratorIdle;

CK_action AR_ACT_GeneratorExplode1;
CK_action AR_ACT_GeneratorExplode2;
CK_action AR_ACT_GeneratorExplode3;
CK_action AR_ACT_GeneratorExplode4;
CK_action AR_ACT_GeneratorExplode5;
CK_action AR_ACT_GeneratorExplode6;


void AR_SpawnGenerator(int tileX, int tileY);
void AR_GeneratorCollide(CK_object *obj1, CK_object *obj2);
void AR_GeneratorExplode(CK_object *obj);
void AR_GeneratorWin(CK_object *obj);

CK_action AR_ACT_GeneratorIdle = {GENERATORIDLESPR, GENERATORIDLESPR, AT_Frame, 0, 0, 60, 0, 0, NULL, AR_GeneratorCollide, CK_BasicDrawFunc1, &AR_ACT_GeneratorIdle};

CK_action AR_ACT_GeneratorExplode1 = {GENERATOREXPLODE1SPR, GENERATOREXPLODE1SPR, AT_UnscaledOnce, 0, 0, 10, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_GeneratorExplode2};
CK_action AR_ACT_GeneratorExplode2 = {GENERATOREXPLODE2SPR, GENERATOREXPLODE2SPR, AT_UnscaledOnce, 0, 0, 10, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_GeneratorExplode3};
CK_action AR_ACT_GeneratorExplode3 = {GENERATOREXPLODE3SPR, GENERATOREXPLODE3SPR, AT_UnscaledOnce, 0, 0, 10, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_GeneratorExplode4};
CK_action AR_ACT_GeneratorExplode4 = {GENERATOREXPLODE4SPR, GENERATOREXPLODE4SPR, AT_UnscaledOnce, 1, 1, 10, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_GeneratorExplode5};
CK_action AR_ACT_GeneratorExplode5 = {GENERATOREXPLODE5SPR, GENERATOREXPLODE5SPR, AT_UnscaledOnce, 1, 1, 10, 0, 0, NULL, NULL, CK_BasicDrawFunc1, &AR_ACT_GeneratorExplode6};
CK_action AR_ACT_GeneratorExplode6 = {GENERATOREXPLODE6SPR, GENERATOREXPLODE6SPR, AT_UnscaledOnce, 1, 1, 10, 0, 0, AR_GeneratorWin, NULL, CK_BasicDrawFunc1, &AR_ACT_GeneratorExplode6};

char ck_generatorMsg1[] = 
	"That Atroxian generator\n"
	"core has been fried!\n"
	"Now to search for\n"
	"the other generator." ;
	
char ck_generatorMsg2[] = 
	"That Atroxian generator\n"
	"core has been fried!\n"
	"Now I can enter the\n"
	"Krile city in the\n"
	"northeast.\n";

char *ck_generatorMsgs[] = {
	ck_generatorMsg1,
	ck_generatorMsg2,
};
void AR_SpawnGenerator(int tileX, int tileY)
{
	CK_object *new_object = CK_GetNewObj(false);
	new_object->type = CT_Generator;
	new_object->active = OBJ_ACTIVE;
	new_object->zLayer = 0;
	new_object->posX = tileX << 8;
	new_object->posY = (tileY << 8) - 0x180;
	new_object->clipped = CLIP_not;
	CK_SetAction(new_object, &AR_ACT_GeneratorIdle);
}

void AR_GeneratorCollide(CK_object *obj1, CK_object *obj2)
{

	if (obj2->type == CT_Stunner)
	{
		CK_ShotHit(obj2);
		CK_SetAction2(obj1, &AR_ACT_GeneratorExplode1);
		SD_PlaySound(GENERATOREXPLODESND);
	}	
}

#if 0
void AR_GeneratorExplode(CK_object *obj)
{
	CK_object *newobj = CK_GetNewObj(true); 
	newobj->posX = obj->clipRects.unitX1 - 0x80 + ((obj->clipRects.unitX2 - obj->clipRects.unitX1) * US_RndT() >> 8);
	newobj->posY = obj->clipRects.unitY1 - 0x80 + ((obj->clipRects.unitY2 - obj->clipRects.unitY1) * US_RndT() >> 8);
	newobj->type = CT_Friendly;
	newobj->clipped = CLIP_not;
	newobj->active = OBJ_EXISTS_ONLY_ONSCREEN;
	newobj->zLayer = 2;
	CK_SetAction(newobj, &AR_ACT_GeneratorExplosion1);
	SD_PlaySound(4);
}
#endif

void AR_GeneratorWin(CK_object *obj)
{
	if (ck_gameState.generatorsDestroyed >= 2)
	{
		ck_levelState = LS_LevelComplete;
	}
	else if (ck_keenObj->currentAction != &CK_ACT_keenDie0 && ck_keenObj->currentAction != &CK_ACT_keenDie1)
	{
		ck_levelState = LS_Generator;
	}
}



void AR_GeneratorWinMessage (void)
{
	SD_WaitSoundDone();
	CA_UpLevel();
	CA_ClearMarks();	// K1n9_Duk3 fix (just in case...)
	grneeded[STARTPICS+105] |= ca_levelbit;
	grneeded[STARTPICS+106] |= ca_levelbit;
	CA_CacheMarks(NULL);
	VW_UpdateScreen();
	StartMusic(-1);

	// Draw Talking pic + msg
	US_CenterWindow(26,8);
	VWB_DrawPic(WindowX, WindowY, 105 + STARTPICS);
	PrintY += 6;
	WindowW -= 48;
	WindowX += 48;
	US_CPrint(ck_generatorMsgs[ck_gameState.generatorsDestroyed++]);
	VW_UpdateScreen();
	VW_WaitVBL(60);
	IN_ClearKeysDown();
	IN_Ack();

	// K1n9_Duk3 fix: stop music before re-caching the graphics to make more memory available
	StopMusic();
	CA_DownLevel();

	// K1n9_Duk3 says: scorebox sprite might have been re-cached
	// by CA_DownLevel, but the level ends after this anyway.
}

char ck_generatorMapMsg[] = 
	"I must destroy\n"
	"the generator that\n"
	"powers this\n"
	"electric barrier.";

void AR_GeneratorMapMessage (void)
{
	SD_WaitSoundDone();
	CA_UpLevel();
	CA_ClearMarks();	// K1n9_Duk3 fix (just in case...)
	grneeded[STARTPICS+105] |= ca_levelbit;
	CA_CacheMarks(NULL);
	VW_UpdateScreen();

	// Draw Talking pic + msg
	US_CenterWindow(26,8);
	VWB_DrawPic(WindowX, WindowY, 105 + STARTPICS);
	PrintY += 6;
	WindowW -= 48;
	WindowX += 48;
	US_CPrint(ck_generatorMapMsg);
	VW_UpdateScreen();
	VW_WaitVBL(20);
	IN_ClearKeysDown();
	IN_Ack();

	CA_DownLevel();

	// K1n9_Duk3 fix: force scorebox to update (in case the sprite was re-cached by CA_DownLevel)
	ck_scoreBoxObj->user1 = -1;
	ck_scoreBoxObj->user2 = -1;
	ck_scoreBoxObj->user3 = -1;
	ck_scoreBoxObj->user4 = -1;
}
//===========================================================================
//
// KRILE ELDER
// 
//===========================================================================

CK_action AR_ACT_ElderIdle;


void AR_SpawnElder(int tileX, int tileY);
void AR_ElderCollide(CK_object *obj1, CK_object *obj2);
void AR_ElderWinMessage(void);

CK_action AR_ACT_ElderIdle = {ELDER2SPR, ELDER2SPR, AT_Frame, 0, 0, 60, 0, 0, NULL, AR_ElderCollide, CK_BasicDrawFunc1, &AR_ACT_ElderIdle};

#if 0
char ck_elderMsg1[] = 
	"Keen, I am the Krile's\n" 
	"elder leader. My people's\n" 
	"fate rests in your hands.\n";

char ck_elderMsg2[] = 
	"Use this orb of light\n" 
	"to brighten the crystal\n" 
	"core and save our world.\n";

char ck_elderMsg3[] =
	"Thanks. I will stop\n" 
	"Atroxus once and for all.\n";

char ck_elderMsg4[] =
	"Use this travel pad to\n" 
	"reach the Sky Sanctum.\n" 
	"Good luck!\n";

char *ck_elderMsgs[] = {
	ck_elderMsg1,
	ck_elderMsg2,
	ck_elderMsg3,
	ck_elderMsg4,
};
#endif

void AR_SpawnElder(int tileX, int tileY)
{
	CK_object *new_object = CK_GetNewObj(false);
	new_object->type = CT_Elder;
	new_object->active = OBJ_ACTIVE;
	new_object->zLayer = 0;
	new_object->posX = tileX << 8;
	new_object->posY = (tileY << 8);
	new_object->clipped = CLIP_not;
	CK_SetAction(new_object, &AR_ACT_ElderIdle);
}

void AR_ElderCollide(CK_object *obj1, CK_object *obj2)
{

	if (obj2->type == CT_Player)
	{
		ck_levelState = LS_Elder;
	}	
}


void AR_ElderWinMessage (void)
{
#if 0
	SD_WaitSoundDone();
	CA_UpLevel();
	CA_ClearMarks();	// K1n9_Duk3 fix (just in case...)
	grneeded[STARTPICS+105] |= ca_levelbit;
	grneeded[STARTPICS+12] |= ca_levelbit;
	CA_CacheMarks(NULL);
	VW_UpdateScreen();
	StartMusic(-1);

	// Draw Talking pic + msg
	US_CenterWindow(26,8);
	VWB_DrawPic(WindowX, WindowY, 12 + STARTPICS);
	PrintY += 6;
	WindowW -= 48;
	WindowX += 48;
	US_CPrint(ck_elderMsgs[0]);
	VW_UpdateScreen();
	VW_WaitVBL(60);
	IN_ClearKeysDown();
	IN_Ack();

	// Draw Talking pic + msg
	US_CenterWindow(26,8);
	VWB_DrawPic(WindowX, WindowY, 12 + STARTPICS);
	PrintY += 6;
	WindowW -= 48;
	WindowX += 48;
	US_CPrint(ck_elderMsgs[1]);
	VW_UpdateScreen();
	VW_WaitVBL(60);
	IN_ClearKeysDown();
	IN_Ack();


	// Draw Talking pic + msg
	US_CenterWindow(26,8);
	VWB_DrawPic(WindowW-48+WindowX, WindowY, 105 + STARTPICS);
	PrintY += 6;
	WindowW -= 48;
	US_CPrint(ck_elderMsgs[2]);
	VW_UpdateScreen();
	VW_WaitVBL(60);
	IN_ClearKeysDown();
	IN_Ack();

	// Draw Talking pic + msg
	US_CenterWindow(26,8);
	VWB_DrawPic(WindowX, WindowY, 12 + STARTPICS);
	PrintY += 6;
	WindowW -= 48;
	WindowX += 48;
	US_CPrint(ck_elderMsgs[3]);
	VW_UpdateScreen();
	VW_WaitVBL(60);
	IN_ClearKeysDown();
	IN_Ack();

	// K1n9_Duk3 fix: stop music before re-caching the graphics to make more memory available
	StopMusic();
	CA_DownLevel();
#endif

	EndScreens(2);
	ck_gameState.endscreen_played[2] = true;

}

//===========================================================================
//
// MAP BRIDGE SWITCH
// 
//===========================================================================


void AR_SpawnBridge(int tileX, int tileY);
void AR_BridgeCollide(CK_object *obj1, CK_object *obj2);
void AR_BridgeWinMessage(void);

CK_action AR_ACT_BridgeIdle = {BRIDGESPR, BRIDGESPR, AT_Frame, 0, 0, 60, 0, 0, NULL, AR_BridgeCollide, CK_BasicDrawFunc1, &AR_ACT_BridgeIdle};

char ck_bridgeMsg[] = 
	"Cool! This activates\n" 
	"Atroxian bridges.\n";


void AR_SpawnBridge(int tileX, int tileY)
{
	CK_object *new_object = CK_GetNewObj(false);
	new_object->type = CT_Bridge;
	new_object->active = OBJ_ACTIVE;
	new_object->zLayer = 0;
	new_object->posX = tileX << 8;
	new_object->posY = (tileY << 8);
	new_object->clipped = CLIP_not;
	CK_SetAction(new_object, &AR_ACT_BridgeIdle);
}

void AR_BridgeCollide(CK_object *obj1, CK_object *obj2)
{

	if (obj2->type == CT_Player)
	{
		ck_levelState = LS_Bridge;
	}	
}

void AR_BridgeWinMessage (void)
{
	SD_WaitSoundDone();
	CA_UpLevel();
	CA_ClearMarks();	// K1n9_Duk3 fix (just in case...)
	grneeded[STARTPICS+105] |= ca_levelbit;
	CA_CacheMarks(NULL);
	VW_UpdateScreen();
	StartMusic(-1);

	// Draw Talking pic + msg
	US_CenterWindow(26,8);
	VWB_DrawPic(WindowX, WindowY, 105 + STARTPICS);
	PrintY += 6;
	WindowW -= 48;
	WindowX += 48;
	US_CPrint(ck_bridgeMsg);
	VW_UpdateScreen();
	VW_WaitVBL(60);
	IN_ClearKeysDown();
	IN_Ack();

	// K1n9_Duk3 fix: stop music before re-caching the graphics to make more memory available
	StopMusic();
	CA_DownLevel();

	// K1n9_Duk3 says: scorebox sprite might have been re-cached
	// by CA_DownLevel, but the level ends after this anyway.
}
//===========================================================================
//
// KRILE BWB ELDER
// 
//===========================================================================

CK_action AR_ACT_ElderBWBIdle;


void AR_SpawnBWBElder(int tileX, int tileY);
void AR_ElderBWBCollide(CK_object *obj1, CK_object *obj2);

CK_action AR_ACT_ElderBWBIdle = {ELDER1SPR, ELDER1SPR, AT_Frame, 0, 0, 60, 0, 0, NULL, AR_ElderBWBCollide, CK_BasicDrawFunc1, &AR_ACT_ElderBWBIdle};


void AR_SpawnBWBElder(int tileX, int tileY)
{
	CK_object *new_object;

	if (ck_gameState.endscreen_played[1])
		return;

	new_object = CK_GetNewObj(false);
	new_object->type = CT_Friendly;
	new_object->active = OBJ_ACTIVE;
	new_object->zLayer = 0;
	new_object->posX = tileX << 8;
	new_object->posY = (tileY << 8);
	new_object->clipped = CLIP_not;
	CK_SetAction(new_object, &AR_ACT_ElderBWBIdle);
}

void AR_ElderBWBCollide(CK_object *obj1, CK_object *obj2)
{
	if (obj2->type != CT_Player)
		return;

	if (ck_gameState.endscreen_played[1])
		return;

	EndScreens(1);
	ck_gameState.endscreen_played[1] = true;
	//RF_ForceRefresh();	// K1n9_Duk3 fix
	CK_RemoveObj(obj1);
	StartMusic(25);

	// Force scorebox redraw if it's enabled
	if (showscorebox)
	{
		ck_scoreBoxObj->user1 = ck_scoreBoxObj->user2 = ck_scoreBoxObj->user3 = ck_scoreBoxObj->user4 = -1;
		// Non-vanilla addition: redraw the scorebox immediately without
		// waiting for a PlayLoop
		CK_UpdateScoreBox(ck_scoreBoxObj);
	}
	RF_ForceRefresh();	// K1n9_Duk3 fix: refresh here to avoid flicker
}

//===========================================================================
//
// Atroxus Pre Battle
// 
//===========================================================================

CK_action AR_ACT_AtroxusDialogueIdle;

void AR_SpawnAtroxusDialogue(int tileX, int tileY);
void AR_AtroxusDialogueCollide(CK_object *obj1, CK_object *obj2);

CK_action AR_ACT_AtroxusDialogueIdle = {ATROXUSDETECTSPR, ATROXUSDETECTSPR, AT_Frame, 0, 0, 60, 0, 0, NULL, AR_AtroxusDialogueCollide, CK_BasicDrawFunc1, &AR_ACT_AtroxusDialogueIdle};


void AR_SpawnAtroxusDialogue(int tileX, int tileY)
{
	CK_object *new_object;

	if (ck_gameState.endscreen_played[3])
		return;

	new_object = CK_GetNewObj(false);
	new_object->type = CT_Friendly;
	new_object->active = OBJ_ACTIVE;
	new_object->zLayer = 0;
	new_object->posX = tileX << 8;
	new_object->posY = (tileY << 8);
	new_object->clipped = CLIP_not;
	CK_SetAction(new_object, &AR_ACT_AtroxusDialogueIdle);
}

void AR_AtroxusDialogueCollide(CK_object *obj1, CK_object *obj2)
{
	if (obj2->type != CT_Player)
		return;

	ck_levelState = LS_Atroxus;
}

//===========================================================================
//
// Secret Krile
// 
//===========================================================================

CK_action AR_ACT_SecretKrileIdle;


void AR_SpawnSecretKrile(int tileX, int tileY);
void AR_SecretKrileCollide(CK_object *obj1, CK_object *obj2);

CK_action AR_ACT_SecretKrileIdle = {ELDER1SPR, ELDER1SPR, AT_Frame, 0, 0, 60, 0, 0, NULL, AR_SecretKrileCollide, CK_BasicDrawFunc1, &AR_ACT_SecretKrileIdle};


void AR_SpawnSecretKrile(int tileX, int tileY)
{
	CK_object *new_object;

	if (ck_gameState.endscreen_played[4])
		return;

	new_object = CK_GetNewObj(false);
	new_object->type = CT_Friendly;
	new_object->active = OBJ_ACTIVE;
	new_object->zLayer = 0;
	new_object->posX = tileX << 8;
	new_object->posY = (tileY << 8);
	new_object->clipped = CLIP_not;
	CK_SetAction(new_object, &AR_ACT_SecretKrileIdle);
}

void AR_SecretKrileCollide(CK_object *obj1, CK_object *obj2)
{
	if (obj2->type != CT_Player)
		return;

	if (ck_gameState.endscreen_played[4])
		return;

	EndScreens(4);
	ck_gameState.endscreen_played[4] = true;
	//RF_ForceRefresh();	// K1n9_Duk3 fix
	CK_RemoveObj(obj1);
	StartMusic(18);

	ck_gameState.doSecretEnding - true;	// K1n9_Duk3 fixme: this can't be right

	// Force scorebox redraw if it's enabled
	if (showscorebox)
	{
		ck_scoreBoxObj->user1 = ck_scoreBoxObj->user2 = ck_scoreBoxObj->user3 = ck_scoreBoxObj->user4 = -1;
		// Non-vanilla addition: redraw the scorebox immediately without
		// waiting for a PlayLoop
		CK_UpdateScoreBox(ck_scoreBoxObj);
	}
	RF_ForceRefresh();	// K1n9_Duk3 fix: refresh here to avoid flicker
}

//===========================================================================
//
// Atroxus Box
// 
//===========================================================================

CK_action AR_ACT_AtroxusBoxIdle;

void AR_SpawnAtroxusBox(int tileX, int tileY);
void AR_AtroxusBoxCollide(CK_object *obj1, CK_object *obj2);

CK_action AR_ACT_AtroxusBoxIdle = {ATROXUSDETECTSPR, ATROXUSDETECTSPR, AT_Frame, 0, 0, 60, 0, 0, NULL, AR_AtroxusBoxCollide, CK_BasicDrawFunc1, &AR_ACT_AtroxusBoxIdle};

int atroxus_endscreens[] = {9, 10, 11};

void AR_SpawnAtroxusBox(int tileX, int tileY)
{
	int i;
	CK_object *new_object;

	if (ck_gameState.atroxusBoxesPlayed >= 3)
		return;

	for (i = 0; i < 3; i++)
		if (ck_gameState.atroxusBoxesLevels[i] == ck_gameState.currentLevel)
			return;

	new_object = CK_GetNewObj(false);
	new_object->type = CT_Friendly;
	new_object->active = OBJ_ACTIVE;
	new_object->zLayer = 0;
	new_object->posX = tileX << 8;
	new_object->posY = (tileY << 8);
	new_object->clipped = CLIP_not;
	CK_SetAction(new_object, &AR_ACT_AtroxusBoxIdle);
}

void AR_AtroxusBoxCollide(CK_object *obj1, CK_object *obj2)
{

	if (obj2->type != CT_Player)
		return;

	ck_gameState.atroxusBoxesLevels[ck_gameState.atroxusBoxesPlayed] = ck_gameState.currentLevel;
	EndScreens(atroxus_endscreens[ck_gameState.atroxusBoxesPlayed++]);

	//RF_ForceRefresh();	// K1n9_Duk3 fix
	CK_RemoveObj(obj1);
	StartMusic(ck_gameState.currentLevel);

	// Force scorebox redraw if it's enabled
	if (showscorebox)
	{
		ck_scoreBoxObj->user1 = ck_scoreBoxObj->user2 = ck_scoreBoxObj->user3 = ck_scoreBoxObj->user4 = -1;
		// Non-vanilla addition: redraw the scorebox immediately without
		// waiting for a PlayLoop
		CK_UpdateScoreBox(ck_scoreBoxObj);
	}
	RF_ForceRefresh();	// K1n9_Duk3 fix: refresh here to avoid flicker
}
