// WL_TEXT.C

#include "KD_DEF.H"
#pragma	hdrstop

/*
=============================================================================

TEXT FORMATTING COMMANDS
------------------------
^C<hex digit>  			Change text color
^E[enter]				End of layout (all pages)
^G<y>,<x>,<pic>[enter]	Draw a graphic and push margins
^P[enter]				start new page, must be the first chars in a layout
^L<x>,<y>[ENTER]		Locate to a specific spot, x in pixels, y in lines

=============================================================================
*/

/*
=============================================================================

						 LOCAL CONSTANTS

=============================================================================
*/

// K1n9_Duk3 additions:
//#define NOPRECACHE	// use this define to cache all in-text pics on the fly


#define BACKCOLOR		0x04


#define WORDLIMIT		80
#define FONTHEIGHT		10
#define	TOPMARGIN		10
#define BOTTOMMARGIN	10
#define LEFTMARGIN		10
#define RIGHTMARGIN		10
#define PICMARGIN		8
#define TEXTROWS		((200-TOPMARGIN-BOTTOMMARGIN)/FONTHEIGHT)
#define	SPACEWIDTH		7
#define SCREENPIXWIDTH	320
#define SCREENMID		(SCREENPIXWIDTH/2)

// ARHACK
// background styles

typedef enum {
	helpbg_black,
	helpbg_info,
	helpbg_red,
} pagetype;

/*
=============================================================================

						 LOCAL VARIABLES

=============================================================================
*/

int			pagenum,numpages;

unsigned	leftmargin[TEXTROWS],rightmargin[TEXTROWS];
char		far *text;
unsigned	rowon;

int			picx,picy,picnum,picdelay;
boolean		layoutdone;
int			ck_helpSelection;
unsigned	ck_helpScreenChunks[] = { T_HELPMAIN, T_HELPCONTROLS, T_HELPSTORY, T_HELPORDER, T_HELPABOUT};

//===========================================================================

//////////////////////////////////////////////////////////////////////////////
// K1n9_Duk3 additions:

#ifdef NOPRECACHE

void near DrawPic(int x, int y, int num)
{
	// if the pic is not cached, then cache it now,
	// draw it and make it purgable again
	if (grsegs[num])
	{
		VWB_DrawPic(x, y, num);
	}
	else
	{
		CA_CacheGrChunk(num);
		VWB_DrawPic(x, y, num);
		MM_SetPurge(&grsegs[num], 3);
		grneeded[num] &= ~ca_levelbit;	//unmark the chunk
	}
}

#define VWB_DrawPic DrawPic

#endif
//////////////////////////////////////////////////////////////////////////////

/*
=====================
=
= RipToEOL
=
=====================
*/

void RipToEOL (void)
{
	while (*text++ != '\n')		// scan to end of line
	;
}


/*
=====================
=
= ParseNumber
=
=====================
*/

int	ParseNumber (void)
{
	char	ch;
	char	num[80],*numptr;

//
// scan until a number is found
//
	ch = *text;
	while (ch < '0' || ch >'9')
		ch = *++text;

//
// copy the number out
//
	numptr = num;
	do
	{
		*numptr++ = ch;
		ch = *++text;
	} while (ch >= '0' && ch <= '9');
	*numptr = 0;

	return atoi (num);
}



/*
=====================
=
= ParsePicCommand
=
= Call with text pointing just after a ^P
= Upon exit text points to the start of next line
=
=====================
*/

void	ParsePicCommand (void)
{
	picy=ParseNumber();
	picx=ParseNumber();
	picnum=ParseNumber();
	RipToEOL ();
}


void	ParseTimedCommand (void)
{
	picy=ParseNumber();
	picx=ParseNumber();
	picnum=ParseNumber();
	picdelay=ParseNumber();
	RipToEOL ();
}


/*
=====================
=
= TimedPicCommand
=
= Call with text pointing just after a ^P
= Upon exit text points to the start of next line
=
=====================
*/

void	TimedPicCommand (void)
{
	ParseTimedCommand ();

//
// update the screen, and wait for time delay
//
	VW_WaitVBL(1);
	VW_ScreenToScreen(bufferofs, displayofs, 40, 200);

//
// wait for time
//
	TimeCount = 0;
	while (TimeCount < picdelay)
	;

//
// draw pic
//
	VWB_DrawPic (picx&~7,picy,picnum);
}


/*
=====================
=
= HandleCommand
=
=====================
*/

void HandleCommand (void)
{
	int	i,margin,top,bottom;
	int	picwidth,picheight,picmid,piccolor;

	switch (toupper(*++text))
	{
	case 'B':
		picy=ParseNumber();
		picx=ParseNumber();
		picwidth=ParseNumber();
		picheight=ParseNumber();
		piccolor=ParseNumber();
		VWB_Bar(picx,picy,picwidth,picheight,piccolor);
		RipToEOL();
		break;
	case ';':		// comment
		RipToEOL();
		break;
	case 'P':		// ^P is start of next page, ^E is end of file
	case 'E':
		layoutdone = true;
		text--;    	// back up to the '^'
		break;

	case 'C':		// ^c<hex digit> changes text color
		i = toupper(*++text);
		if (i>='0' && i<='9')
			fontcolor = i-'0';
		else if (i>='A' && i<='F')
			fontcolor = i-'A'+10;

#if 0
		fontcolor *= 16;
		i = toupper(*++text);
		if (i>='0' && i<='9')
			fontcolor += i-'0';
		else if (i>='A' && i<='F')
			fontcolor += i-'A'+10;
#endif

		fontcolor ^= BACKCOLOR;
		text++;
		break;

	case '>':		// Indent halfway
		px = 160;
		text++;
		break;

	case 'L':
		py=ParseNumber();
		rowon = (py-TOPMARGIN)/FONTHEIGHT;
		py = TOPMARGIN+rowon*FONTHEIGHT;
		px=ParseNumber();
		while (*text++ != '\n')		// scan to end of line
		;
		break;

	case 'T':		// ^Tyyy,xxx,ppp,ttt waits ttt tics, then draws pic
		TimedPicCommand ();
		break;

	case 'G':		// ^Gyyy,xxx,ppp draws graphic
		ParsePicCommand ();
		VWB_DrawPic (picx&~7,picy,picnum);
		picwidth = pictable[picnum-STARTPICS].width*8;
		picheight = pictable[picnum-STARTPICS].height;
		//
		// adjust margins
		//
		picmid = picx + picwidth/2;
		if (picmid > SCREENMID)
			margin = picx-PICMARGIN;			// new right margin
		else
			margin = picx+picwidth+PICMARGIN;	// new left margin

		top = (picy-TOPMARGIN)/FONTHEIGHT;
		if (top<0)
			top = 0;
		bottom = (picy+picheight-TOPMARGIN)/FONTHEIGHT;
		if (bottom>=TEXTROWS)
			bottom = TEXTROWS-1;

		for (i=top;i<=bottom;i++)
			if (picmid > SCREENMID)
				rightmargin[i] = margin;
			else
				leftmargin[i] = margin;

		//
		// adjust this line if needed
		//
		if (px < leftmargin[rowon])
			px = leftmargin[rowon];
		break;
//////////////////////////////////////////////////////////////////////////////
// K1n9_Duk3 additions:

#if 0	//change this to 1 to use this code

	case 'M':		// ^Mxxx -- Play Music for level xxx
		picnum = (*text == '-')? -ParseNumber() : ParseNumber();
		RipToEOL();
		StartMusic(picnum);
		break;
	case 'S':		// ^Sxxx -- Play sound number xxx
		picnum = ParseNumber();
		RipToEOL();
		SD_PlaySound(picnum);
		break;
	case 'W':		// Wait until sound is done
		SD_WaitSoundDone();
		break;
	case 'D':		// ^Dttt waits ttt tics
		picdelay = ParseNumber();
		RipToEOL();
		VW_WaitVBL(1);
		VW_ScreenToScreen(bufferofs, displayofs, 40, 200);
		TimeCount = 0;
		while (TimeCount < picdelay)
		;
		break;
	case 'Q':		// Quiet - stop sound and music
		StopMusic();
		SD_StopSound();
		break;
#endif
//////////////////////////////////////////////////////////////////////////////
	}
}


/*
=====================
=
= NewLine
=
=====================
*/

void NewLine (void)
{
	char	ch;

	if (++rowon == TEXTROWS)
	{
	//
	// overflowed the page, so skip until next page break
	//
		layoutdone = true;
		do
		{
			if (*text == '^')
			{
				ch = toupper(*(text+1));
				if (ch == 'E' || ch == 'P')
				{
					layoutdone = true;
					return;
				}
			}
			text++;

		} while (1);

	}
	px = leftmargin[rowon];
	py+= FONTHEIGHT;
}



/*
=====================
=
= HandleCtrls
=
=====================
*/

void HandleCtrls (void)
{
	char	ch;

	ch = *text++;			// get the character and advance

	if (ch == '\n')
	{
		NewLine ();
		return;
	}

}


/*
=====================
=
= HandleWord
=
=====================
*/

void HandleWord (void)
{
	char		word[WORDLIMIT];
	int			i,wordindex;
	unsigned	wwidth,wheight,newpos;


	//
	// copy the next word into [word]
	//
	word[0] = *text++;
	wordindex = 1;
	while (*text>32)
	{
		word[wordindex] = *text++;
		if (++wordindex == WORDLIMIT)
			Quit ("PageLayout: Word limit exceeded");
	}
	word[wordindex] = 0;		// stick a null at end for C

	//
	// see if it fits on this line
	//
	VW_MeasurePropString (word,&wwidth,&wheight);

	while (px+wwidth > rightmargin[rowon])
	{
		NewLine ();
		if (layoutdone)
			return;		// overflowed page
	}

	//
	// print it
	//
	newpos = px+wwidth;
	VWB_DrawPropString (word);
	px = newpos;

	//
	// suck up any extra spaces
	//
	while (*text == ' ')
	{
		px += SPACEWIDTH;
		text++;
	}
}

/*
=====================
=
= PageLayout
=
= Clears the screen, draws the pics on the page, and word wraps the text.
= Returns a pointer to the terminating command
=
=====================
*/

void PageLayout (pagetype pagestyle)
{
	int		i,oldfontcolor;
	char	ch;

	oldfontcolor = fontcolor;

	fontcolor = 10;

//
// clear the screen
//

	// AR Hack: If not showing number, then we're drawing a black background
	switch (pagestyle)
	{

		case helpbg_red:
		case helpbg_info:
			VWB_Bar (0,0,320,200,BACKCOLOR);
			VWB_DrawPic (0,0,H_TOPWINDOWPIC);
			VWB_DrawPic (0,8,H_LEFTWINDOWPIC);
			VWB_DrawPic (312,8,H_RIGHTWINDOWPIC);

			if (pagestyle == helpbg_info)
				VWB_DrawPic (8,176,H_BOTTOMINFOPIC);
			else
				VWB_DrawPic (8, 192, H_BOTTOMWINDOWPIC);

			break;

		case helpbg_black:
			VWB_Bar (0,0,320,200,BLACK);
			break;
	}


	for (i=0;i<TEXTROWS;i++)
	{
		leftmargin[i] = LEFTMARGIN;
		rightmargin[i] = SCREENPIXWIDTH-RIGHTMARGIN;
	}

	px = LEFTMARGIN;
	py = TOPMARGIN;
	rowon = 0;
	layoutdone = false;

//
// make sure we are starting layout text (^P first command)
//
	while (*text <= 32)
		text++;

	if (*text != '^' && toupper(*++text) != 'P')
		Quit ("PageLayout: Text not headed with ^P");

	while (*text++ != '\n')
	;


//
// process text stream
//
	do
	{
		ch = *text;

		if (ch == '^')
			HandleCommand ();
		else
		if (ch == 9)
		{
		//Tab
		 px = (px+8)&0xf8;
		 text++;
		}
		else if (ch <= 32)
			HandleCtrls ();
		else
			HandleWord ();

	} while (!layoutdone);

	pagenum++;

	if (pagestyle == helpbg_info)
	{
		strcpy (str,"pg ");
		itoa (pagenum,str2,10);
		strcat (str,str2);
		strcat (str," of ");
		itoa (numpages,str2,10);
		strcat (str,str2);

		py = 186;
		px = 218;
		fontcolor = 12^BACKCOLOR;

		VWB_DrawPropString (str);
	}

	fontcolor = oldfontcolor;
}

//===========================================================================

/*
=====================
=
= BackPage
=
= Scans for a previous ^P
=
=====================
*/

void BackPage (void)
{
	pagenum--;
	do
	{
		text--;
		if (*text == '^' && toupper(*(text+1)) == 'P')
			return;
	} while (1);
}


//===========================================================================


/*
=====================
=
= CacheLayoutGraphics
=
= Scans an entire layout file (until a ^E) marking all graphics used, and
= counting pages, then caches the graphics in
= AR HACK: 
=
=====================
*/
void CacheLayoutGraphics (pagetype pagestyle)
{
	char	far *bombpoint, far *textstart;
	char	ch;

	textstart = text;
	bombpoint = text+30000;
	numpages = pagenum = 0;

	CA_ClearMarks();	// K1n9_Duk3 addition (only cache what gets marked below)

	do
	{
		if (*text == '^')
		{
			ch = toupper(*++text);
			if (ch == 'P')		// start of a page
				numpages++;
			if (ch == 'E')		// end of file, so load graphics and return
			{
				switch (pagestyle)
				{
					case helpbg_black:
						break;

					case helpbg_info:
					case helpbg_red:
						CA_MarkGrChunk(H_TOPWINDOWPIC);
						CA_MarkGrChunk(H_LEFTWINDOWPIC);
						CA_MarkGrChunk(H_RIGHTWINDOWPIC);

					if (pagestyle == helpbg_info)
						CA_MarkGrChunk(H_BOTTOMINFOPIC);
					else
						CA_MarkGrChunk(H_BOTTOMWINDOWPIC);
					break;
				}
				text = textstart;
				CA_CacheMarks (NULL);
				return;
			}
			if (ch == 'G')		// draw graphic command, so mark graphics
			{
				ParsePicCommand ();
#ifndef NOPRECACHE
				CA_MarkGrChunk (picnum);
#endif
			}
			if (ch == 'T')		// timed draw graphic command, so mark graphics
			{
				ParseTimedCommand ();
#ifndef NOPRECACHE
				CA_MarkGrChunk (picnum);
#endif
			}
		}
		else
			text++;

	} while (text<bombpoint);

	Quit ("CacheLayoutGraphics: No ^E to terminate file!");
}


/*
=====================
=
= ShowHelp
=
=====================
*/

int ShowHelp ()
{

	ControlInfo control; 
	CursorInfo cursor;
	int si;

	// Draw the backdrop
	VWB_Bar(0,0,320,200, BACKCOLOR);

	CA_SetGrPurge();	// K1n9_Duk3 addition (clear graphics cache)

	// Load border graphics
	CA_CacheGrChunk(H_MENUPIC);
	CA_CacheGrChunk(H_HANDPIC);
	CA_CacheGrChunk(H_TOPWINDOWPIC);
	CA_CacheGrChunk(H_LEFTWINDOWPIC);
	CA_CacheGrChunk(H_RIGHTWINDOWPIC);
	CA_CacheGrChunk(H_BOTTOMWINDOWPIC);

	// Draw border graphics
	VWB_DrawPic(0,0,H_TOPWINDOWPIC);
	VWB_DrawPic(0,8,H_LEFTWINDOWPIC);
	VWB_DrawPic(312,8,H_RIGHTWINDOWPIC);
	VWB_DrawPic(8,192,H_BOTTOMWINDOWPIC);
	VWB_DrawPic(96,8,H_MENUPIC);

	si = 0;
	IN_ClearKeysDown();

	// Move Hand up and down
	while (true)
	{
		if (ck_helpSelection < 0)
			ck_helpSelection = 0;
		else if (ck_helpSelection > 4)
			ck_helpSelection = 4;

		// Draw the pointer
		VWB_DrawPic(48, 48 + 24 * ck_helpSelection, H_HANDPIC);
		VW_UpdateScreen();

		// Erase the pointer
		VWB_Bar(48, 48 + 24 * ck_helpSelection, 39, 24, BACKCOLOR);

		IN_ReadControl(0, &control);
		IN_ReadCursor(&cursor);

		if (LastScan)
		{
				int di = LastScan;
				IN_ClearKeysDown();
				if (di == sc_UpArrow)
					ck_helpSelection--;
				else if (di == sc_DownArrow)
					ck_helpSelection++;
				else if (di == sc_Escape)
				{
					VW_ClearVideo(BACKCOLOR);
					return -1;
				}
				else if (di == sc_Enter)
				{
					VW_ClearVideo(BACKCOLOR);
					return ck_helpSelection;
				}
		}


		si += cursor.y;

		// Select a help topic
		if (cursor.button0 || cursor.button1 || control.button0 || control.button1)
		{
			VW_ClearVideo(BACKCOLOR);
			return ck_helpSelection;
		}


		// Move down
		else if (si < -40)
		{
			si += 40;
			ck_helpSelection--;
		}

		// Move up
		else if (si > 40)
		{
			si -= 40;
			ck_helpSelection++;
		}
	}
				

}


//===========================================================================


/*
=================
=
= HelpScreens
=
=================
*/
void HelpScreens (void)
{
	int			artnum;
	memptr		layout;
	int			oldfont;
	unsigned	olddisplayofs, oldbufferofs;
	int 		helppage;


	oldfont = fontnumber;
	olddisplayofs = displayofs;
	oldbufferofs = bufferofs;
	fontnumber = 0;
	EGAMAPMASK(0xF);

	CA_UpLevel();
	CA_SetGrPurge();
	VW_ClearVideo(BACKCOLOR);
	RF_FixOfs();

	bufferofs = 0;
	displayofs = 0x8000;
	VW_SetScreen(displayofs, 0);

	StartMusic(HELP_MUS, true);

	//loop start
	while (1)
	{
		boolean refresh;

		// Choose which help screen to draw
		helppage = ShowHelp();
		VW_ClearVideo(BACKCOLOR);

		if (helppage == -1)
		{
			// Exit

			// K1n9_Duk3 fix: stop the music before re-caching to make more memory available
			StopMusic();
			CA_ClearMarks();

			CA_DownLevel();
			IN_ClearKeysDown();
			bufferofs = oldbufferofs;
			displayofs = olddisplayofs;
			fontnumber = oldfont;
			VW_ClearVideo(BACKCOLOR);
			RF_FixOfs();
			return;
		}

		artnum = ck_helpScreenChunks[helppage];
		CA_CacheGrChunk(artnum);

		// K1n9_Duk3 fix: prevent Memory Manager from messing with the text
		MM_SetLock(&grsegs[artnum], true);

		text = grsegs[artnum];
		CacheLayoutGraphics(helpbg_info);

		refresh = true;

		// Flip pages
		while (1)
		{
			if (refresh)
			{
				unsigned buff;

				refresh = false;
				PageLayout(helpbg_info);

				// Display and flip pages
				VW_SetScreen(bufferofs, 0);
				buff = displayofs;
				displayofs = bufferofs;
				bufferofs = buff;

			}

			// Wait for input
			for (LastScan = sc_None; !LastScan;)
				;

			switch(LastScan)
			{
				case sc_UpArrow:
				case sc_PgUp:
				case sc_LeftArrow:
					if (pagenum > 1)
					{
						BackPage();
						BackPage();
						refresh = true;
					}
					break;

				case sc_DownArrow:
				case sc_PgDn:
				case sc_RightArrow:
					if (pagenum < numpages)
					{
						refresh = true;
					}
					break;
			}

			if (LastScan == sc_Escape)
			{
				// K1n9_Duk3 fix: unlock the text (probably not needed here)
				MM_SetLock(&grsegs[artnum], false);

				// Go back to help selection menu
				MM_FreePtr(&grsegs[artnum]);
				IN_ClearKeysDown();
				break;
			}
		}
	}

}

int endscreen_chunks[] = {
	T_ENDDIALOGUE, T_BWBKRILE, T_KRILEELDER, T_ATROXUS, T_SECRETKRILE, 
	T_SECRETEND, T_END2, T_END3, T_NORMALEND, T_ATROXUSBOX1, 
	T_ATROXUSBOX2, T_ATROXUSBOX3
};

int endscreen_songs[] = {
	ENDING_MUS, KRILE_THEME_MUS, ELDER_TEXT_MUS, ATROXUS_TEXT_MUS, KRILE_THEME_MUS, 
	ENDING_MUS, ENDING_MUS, ENDING_MUS, NORMAL_END_MUS, ATROXUS_BOX_MUS, 
	ATROXUS_BOX_MUS, ATROXUS_BOX_MUS };

int endscreen_type[] = {
	helpbg_black, helpbg_red, helpbg_red, helpbg_black, helpbg_red, 
	helpbg_red, helpbg_black, helpbg_black, helpbg_red, helpbg_black, 
	helpbg_black, helpbg_black
};

void EndScreens(int dialogue)
{

	// Don't interrupt the ending music
	if (dialogue != 6)
		StopMusic();
	VW_FadeOut();

	VW_ClearVideo(0);

	RF_FixOfs();
	CA_UpLevel();

	// Clear out the main menu graphics
	CA_ClearMarks();


	CA_SetGrPurge();

	// Do level dependent caching
	//
	

	CA_CacheGrChunk(endscreen_chunks[dialogue]);

	// K1n9_Duk3 fix: prevent Memory Manager from messing with the text
	MM_SetLock(&grsegs[endscreen_chunks[dialogue]], true);

	text = grsegs[endscreen_chunks[dialogue]];
	CacheLayoutGraphics(endscreen_type[dialogue]);

	// Cache blinking arrow graphics	(K1n9_Duk3 fix: moved down here)
	CA_CacheGrChunk(47);
	CA_CacheGrChunk(48);
	
	if (dialogue != 6)
		StartMusic(endscreen_songs[dialogue]);
	VW_FadeIn();

	while (pagenum < numpages)
	{

		PageLayout(endscreen_type[dialogue]);
		IN_ClearKeysDown();
		VW_SetScreen(bufferofs, 0);

		// Flash the arrow at the bottom right
		do
		{
			int si;

			VWB_DrawPic(298, 184, 47);
			for (si = 0; si < 70; si++)
			{
				if (IN_IsUserInput())
					goto nextpage;

				VW_WaitVBL(1);
			}

			VWB_DrawPic(298, 184, 48);
			for (si = 0; si < 70; si++)
			{
				if (IN_IsUserInput())
				{
					if (LastScan == sc_Escape)
						goto done;
					else
						goto nextpage;
				}

				VW_WaitVBL(1);
			}


		} while (1);

nextpage:
	}
done:

	// K1n9_Duk3 fix: unlock the text (definitely needed here!)
	MM_SetLock(&grsegs[endscreen_chunks[dialogue]], false);

	// Clear out the endscreen graphics 
	CA_ClearMarks();

	if (dialogue != 0)
		StopMusic();

	CA_DownLevel();
	IN_ClearKeysDown();
	VW_ClearVideo(0);
	RF_FixOfs();
}
